# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('../'))

# -- Project information -----------------------------------------------------

project = 'PynPoint'
copyright = '2014-2023, Tomas Stolker, Markus Bonse, Sascha Quanz, and Adam Amara'
author = 'Tomas Stolker, Markus Bonse, Sascha Quanz, and Adam Amara'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'nbsphinx'
]

# Disable notebook timeout
nbsphinx_timeout = -1

# Allow errors from notebooks
nbsphinx_allow_errors = True

autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build',
                    'Thumbs.db',
                    '.DS_Store',
                    'tutorials/.ipynb_checkpoints/*']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'path_to_docs': 'docs',
    'repository_url': 'https://github.com/PynPoint/PynPoint',
    'repository_branch': 'main',
    'launch_buttons': {
        'binderhub_url': 'https://mybinder.org',
        'notebook_interface': 'jupyterlab',
    },
    'use_edit_page_button': True,
    'use_issues_button': True,
    'use_repository_button': True,
    'use_download_button': True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

html_logo = '_static/logo.png'
html_favicon = '_static/favicon.png'
html_search_language = 'en'

html_context = {'display_github': True,
                'github_user': 'PynPoint',
                'github_repo': 'PynPoint',
                'github_version': 'main/docs/'}



import os

import pytest
import h5py
import numpy as np

from pynpoint.core.dataio import DataStorage


class TestDataStorage:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_data = os.path.dirname(__file__) + '/PynPoint_database.hdf5'

    def test_create_storage_with_existing_database(self) -> None:

        np.random.seed(1)
        images = np.random.normal(loc=0, scale=2e-4, size=(10, 100, 100))

        with h5py.File(self.test_data, 'w') as hdf_file:
            hdf_file.create_dataset('images', data=images)

        storage = DataStorage(self.test_data)
        storage.open_connection()
        data = storage.m_data_bank['images']

        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)
        assert np.mean(data) == pytest.approx(1.0506056979365338e-06, rel=self.limit, abs=0.)

        os.remove(self.test_data)

    def test_create_storage_without_existing_database(self) -> None:

        storage = DataStorage(self.test_data)
        storage.open_connection()
        storage.m_data_bank['data'] = [0, 1, 2, 5, 7]

        assert storage.m_data_bank['data'][2] == 2
        assert list(storage.m_data_bank.keys()) == ['data', ]

        storage.close_connection()

        os.remove(self.test_data)

    def test_create_storage_with_wrong_location(self) -> None:

        file_in = '/test/test.hdf5'

        with pytest.raises(AssertionError):
            DataStorage(file_in)

    def test_open_close_connection(self) -> None:

        storage = DataStorage(self.test_data)

        storage.open_connection()
        assert storage.m_open is True

        storage.open_connection()
        assert storage.m_open is True

        storage.close_connection()
        assert storage.m_open is False

        storage.close_connection()
        assert storage.m_open is False

        os.remove(self.test_data)



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.textwriting import TextWritingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestTextWriting:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir, nimages=1)
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=['image.dat', 'data.dat'])

    def test_input_data(self) -> None:

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.0008557279524431129, rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

    def test_text_writing(self) -> None:

        text_write = TextWritingModule(name_in='text_write',
                                       data_tag='images',
                                       file_name='image.dat',
                                       output_dir=None,
                                       header=None)

        self.pipeline.add_module(text_write)
        self.pipeline.run_module('text_write')

        data = np.loadtxt(self.test_dir+'image.dat')

        assert np.sum(data) == pytest.approx(0.0008557279524431129, rel=self.limit, abs=0.)
        assert data.shape == (11, 11)

    def test_text_writing_ndim(self) -> None:

        data_4d = np.random.normal(loc=0, scale=2e-4, size=(5, 5, 5, 5))

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file.create_dataset('data_4d', data=data_4d)

        text_write = TextWritingModule(name_in='write_4d',
                                       data_tag='data_4d',
                                       file_name='data.dat',
                                       output_dir=None,
                                       header=None)

        self.pipeline.add_module(text_write)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('write_4d')

        assert str(error.value) == 'Only 1D or 2D arrays can be written to a text file.'

    def test_text_writing_int(self) -> None:

        data_int = np.arange(1, 11, 1)

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file.create_dataset('data_int', data=data_int)

        text_write = TextWritingModule(name_in='write_int',
                                       data_tag='data_int',
                                       file_name='data.dat',
                                       output_dir=None,
                                       header=None)

        self.pipeline.add_module(text_write)
        self.pipeline.run_module('write_int')

        data = np.loadtxt(self.test_dir+'data.dat')

        assert data == pytest.approx(data_int, rel=self.limit, abs=0.)
        assert data.shape == (10, )



"""
Pipeline modules for aligning and centering of the star.
"""

import math
import time
import warnings

from typing import Optional, Tuple, Union

import numpy as np

from astropy.modeling import fitting, models
from scipy.ndimage import gaussian_filter
from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.image import center_pixel, crop_image, pixel_distance, shift_image, \
                                subpixel_distance
from pynpoint.util.module import memory_frames, progress
from pynpoint.util.apply_func import align_image, apply_shift, fit_2d_function


class StarAlignmentModule(ProcessingModule):
    """
    Pipeline module to align the images with a cross-correlation in Fourier space.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 ref_image_in_tag: Optional[str] = None,
                 interpolation: str = 'spline',
                 accuracy: float = 10.,
                 resize: Optional[float] = None,
                 num_references: int = 10,
                 subframe: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the stack of images that is read as input.
        ref_image_in_tag : str, None
            Tag of the database entry with the reference image(s) that are read as input. If it is
            set to None, a random subsample of *num_references* elements of *image_in_tag* is taken
            as reference images.
        image_out_tag : str
            Tag of the database entry with the images that are written as output.
        interpolation : str
            Type of interpolation that is used for shifting the images (spline, bilinear, or fft).
        accuracy : float
            Upsampling factor for the cross-correlation. Images will be registered to within
            1/accuracy of a pixel.
        resize : float, None
            Scaling factor for the up/down-sampling before the images are shifted. Not used if set
            to None.
        num_references : int
            Number of reference images for the cross-correlation.
        subframe : float, None
            Size (arcsec) of the subframe around the image center that is used for the
            cross-correlation. The full image is used if set to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        if ref_image_in_tag is not None:
            self.m_ref_image_in_port = self.add_input_port(ref_image_in_tag)
        else:
            self.m_ref_image_in_port = None

        self.m_interpolation = interpolation
        self.m_accuracy = accuracy
        self.m_resize = resize
        self.m_num_references = num_references
        self.m_subframe = subframe

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Applies a cross-correlation of the input images with respect to
        a stack of reference images, rescales the image dimensions, and shifts the images to a
        common center.

        Returns
        -------
        NoneType
            None
        """

        if self.m_ref_image_in_port is None:
            random = np.random.choice(self.m_image_in_port.get_shape()[0],
                                      self.m_num_references,
                                      replace=False)

            ref_images = self.m_image_in_port[np.sort(random), :, :]

        else:
            n_ref = self.m_ref_image_in_port.get_shape()[0]

            if n_ref < self.m_num_references:
                warnings.warn(f'Number of available images ({n_ref}) is smaller than '
                              f'num_references ({self.m_num_references}). Using all '
                              f'available images instead.')

                self.m_num_references = n_ref

            ref_index = np.sort(np.random.choice(n_ref, self.m_num_references, replace=False))
            ref_images = self.m_ref_image_in_port[ref_index, :, :]

        if self.m_subframe is not None:
            pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
            self.m_subframe = int(self.m_subframe/pixscale)

        self.apply_function_to_images(align_image,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Aligning images',
                                      func_args=(self.m_interpolation,
                                                 self.m_accuracy,
                                                 self.m_resize,
                                                 self.m_num_references,
                                                 self.m_subframe,
                                                 ref_images.reshape(-1),
                                                 ref_images.shape))

        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        if self.m_resize is not None:
            pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
            new_pixscale = pixscale/self.m_resize
            self.m_image_out_port.add_attribute('PIXSCALE', new_pixscale)
            print(f'New pixel scale (arcsec) = {new_pixscale:.2f}')

        history = f'resize = {self.m_resize}'
        self.m_image_out_port.add_history('StarAlignmentModule', history)
        self.m_image_out_port.close_port()


class FitCenterModule(ProcessingModule):
    """
    Pipeline module for fitting the PSF with a 2D Gaussian or Moffat function.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 fit_out_tag: str,
                 mask_out_tag: Optional[str] = None,
                 method: str = 'full',
                 mask_radii: Tuple[Optional[float], float] = (None, 0.1),
                 sign: str = 'positive',
                 model: str = 'gaussian',
                 filter_size: Optional[float] = None,
                 **kwargs: Union[Tuple[float, float, float, float, float, float, float],
                                 Tuple[float, float, float, float, float, float, float, float],
                                 float]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Database tag of the images that are read as input.
        fit_out_tag : str
            Database tag where the best-fit results and 1σ errors will be stored.
            The data are written in the following format: x offset (pix), x offset error (pix)
            y offset (pix), y offset error (pix), FWHM major axis (arcsec), FWHM major axis error
            (arcsec), FWHM minor axis (arcsec), FWHM minor axis error (arcsec), amplitude (ADU),
            amplitude error (ADU), angle (deg), angle error (deg) measured in counterclockwise
            direction with respect to the upward direction (i.e. east of north), offset (ADU),
            offset error (ADU), power index (only for Moffat function), and power index error
            (only for Moffat function). The ``fit_out_tag`` can be used as argument of ``shift_xy``
            when running the :class:`~pynpoint.processing.centering.ShiftImagesModule`.
        mask_out_tag : str, None
            Database tag where the masked images will be stored. The unmasked part of the images is
            used for the fit. The effect of the smoothing that is applied by setting the ``fwhm``
            argument is also visible in the data of the ``mask_out_tag``. The data are not stored
            if the argument is set to None. The :class:`~pynpoint.core.dataio.OutputPort` of
            ``mask_out_tag`` can only be used when ``CPU = 1``.
        method : str
            Fit and shift each image individually ('full') or only fit the mean of the cube and
            shift each image by this constant offset ('mean'). The 'mean' method can be used in
            case the images are already aligned with
            :class:`~pynpoint.processing.centering.StarAlignmentModule`.
        mask_radii : tuple(float, float), tuple(None, float)
            Inner and outer radius (arcsec) within and beyond which pixels are neglected during the
            fit. The radii are centered at the position that specified with the argument of
            ``guess``, which is the center of the image by default. The outer mask (second value
            of ``mask_radii``) is mandatory whereas radius of the inner mask is optional and can
            be set to None.
        sign : str
            Fit a 'positive' or 'negative' Gaussian/Moffat function. A 'negative' model can be used
            to center coronagraphic data in which a dark hole.
        model : str
            Type of 2D model that is used for the fit ('gaussian' or 'moffat'). Both models are
            elliptical in shape.
        filter_size : float, None
            Standard deviation (arcsec) of the Gaussian filter that is used to smooth the
            images before fitting the model. No filter is applied if set to None.

        Keyword arguments
        -----------------
        guess : tuple(float, float, float, float, float, float, float, float),
                tuple(float, float, float, float, float, float, float, float, float)
            The initial parameter values for the least squares fit: x offset with respect to center
            (pix), y offset with respect to center (pix), FWHM x (pix), FWHM y (pix), amplitude
            (ADU), angle (deg), offset (ADU), and power index (only for Moffat function).

        Returns
        -------
        NoneType
            None
        """

        if 'guess' in kwargs:
            self.m_guess = kwargs['guess']

        else:
            if model == 'gaussian':
                self.m_guess = (0., 0., 1., 1., 1., 0., 0.)

            elif model == 'moffat':
                self.m_guess = (0., 0., 1., 1., 1., 0., 0., 1.)

        if 'radius' in kwargs:
            mask_radii = (None, kwargs['radius'])

            warnings.warn(f'The \'radius\' parameter has been deprecated. Please use the '
                          f'\'mask_radii\' parameter instead. The argument of \'mask_radii\' '
                          f'is set to {mask_radii}.', DeprecationWarning)

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_fit_out_port = self.add_output_port(fit_out_tag)

        if mask_out_tag is None:
            self.m_mask_out_port = None
        else:
            self.m_mask_out_port = self.add_output_port(mask_out_tag)

        self.m_method = method
        self.m_mask_radii = mask_radii
        self.m_sign = sign
        self.m_model = model
        self.m_filter_size = filter_size

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Uses a non-linear least squares (Levenberg-Marquardt) method
        to fit the the individual images or the mean of all images with a 2D Gaussian or Moffat
        function. The best-fit results and errors are stored and contain zeros in case the
        algorithm could not converge. The ``fit_out_tag`` can be used as argument of ``shift_xy``
        when running the :class:`~pynpoint.processing.centering.ShiftImagesModule`.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        cpu = self._m_config_port.get_attribute('CPU')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        if cpu > 1 and self.m_mask_out_port is not None:
            warnings.warn('The mask_out_port can only be used if CPU=1. No data will be '
                          'stored to this output port.')

            del self._m_output_ports[self.m_mask_out_port.tag]
            self.m_mask_out_port = None

        if self.m_mask_radii[0] is None:
            # Convert from arcsec to pixels and change None to 0
            self.m_mask_radii = (0., self.m_mask_radii[1]/pixscale)

        else:
            # Convert from arcsec to pixels
            self.m_mask_radii = (self.m_mask_radii[0]/pixscale, self.m_mask_radii[1]/pixscale)

        if self.m_filter_size:
            # Convert from arcsec to pixels
            self.m_filter_size /= pixscale

        _, xx_grid, yy_grid = pixel_distance(self.m_image_in_port.get_shape()[-2:], position=None)

        rr_ap = subpixel_distance(self.m_image_in_port.get_shape()[-2:],
                                  position=(self.m_guess[1], self.m_guess[0]),
                                  shift_center=False)  # (y, x)

        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        if self.m_method == 'full':

            self.apply_function_to_images(fit_2d_function,
                                          self.m_image_in_port,
                                          self.m_fit_out_port,
                                          'Fitting the stellar PSF',
                                          func_args=(self.m_mask_radii,
                                                     self.m_sign,
                                                     self.m_model,
                                                     self.m_filter_size,
                                                     self.m_guess,
                                                     self.m_mask_out_port,
                                                     xx_grid,
                                                     yy_grid,
                                                     rr_ap,
                                                     pixscale))

        elif self.m_method == 'mean':
            print('Fitting the stellar PSF...', end='')

            im_mean = np.zeros(self.m_image_in_port.get_shape()[1:3])

            for i, _ in enumerate(frames[:-1]):
                im_mean += np.sum(self.m_image_in_port[frames[i]:frames[i+1], ], axis=0)

            best_fit = fit_2d_function(im_mean/float(nimages),
                                       0,
                                       self.m_mask_radii,
                                       self.m_sign,
                                       self.m_model,
                                       self.m_filter_size,
                                       self.m_guess,
                                       self.m_mask_out_port,
                                       xx_grid,
                                       yy_grid,
                                       rr_ap,
                                       pixscale)

            best_fit = best_fit[np.newaxis, ...]
            best_fit = np.repeat(best_fit, nimages, axis=0)

            self.m_fit_out_port.set_all(best_fit, data_dim=2)

            print(' [DONE]')

        history = f'model = {self.m_model}'

        self.m_fit_out_port.copy_attributes(self.m_image_in_port)
        self.m_fit_out_port.add_history('FitCenterModule', history)

        if self.m_mask_out_port:
            self.m_mask_out_port.copy_attributes(self.m_image_in_port)
            self.m_mask_out_port.add_history('FitCenterModule', history)

        self.m_fit_out_port.close_port()


class ShiftImagesModule(ProcessingModule):
    """
    Pipeline module for shifting a stack of images.
    """

    __author__ = 'Tomas Stolker, Benedikt Schmidhuber'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 shift_xy: Union[Tuple[float, float], str],
                 interpolation: str = 'spline') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        shift_xy : tuple(float, float), str
            The shift (pix) in x and y direction as (delta_x, delta_y). Or, a database tag with
            the fit results from the :class:`~pynpoint.processing.centering.FitCenterModule`.
        interpolation : str
            Interpolation type for shifting of the images ('spline', 'bilinear', or 'fft').

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_interpolation = interpolation

        if isinstance(shift_xy, str):
            self.m_fit_in_port = self.add_input_port(shift_xy)
            self.m_shift = None

        else:
            self.m_fit_in_port = None
            self.m_shift = (shift_xy[1], shift_xy[0])

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Shifts an image with a fifth order spline, bilinear, or a
        Fourier shift interpolation.

        Returns
        -------
        NoneType
            None
        """

        constant = True

        # read the fit results from the self.m_fit_in_port if available
        if self.m_fit_in_port is not None:

            self.m_shift = -1.*self.m_fit_in_port[:, [0, 2]]  # (x, y)
            self.m_shift = self.m_shift[:, [1, 0]]  # (y, x)

            # check if data in self.m_fit_in_port is constant for all images using the
            # constant flag
            if not np.allclose(self.m_fit_in_port.get_all() - self.m_fit_in_port[0, ], 0.0):
                constant = False

            if constant:
                # if the offset is constant then use the first element for all images
                self.m_shift = self.m_shift[0, ]

            else:
                # if the offset is not constant, then apply the shifts to each frame individually
                for i, shift in enumerate(self.m_shift):
                    shifted_image = shift_image(self.m_image_in_port[i, ],
                                                shift,
                                                self.m_interpolation)

                    # append the shifted images to the selt.m_image_out_port database entry
                    self.m_image_out_port.append(shifted_image, data_dim=3)

                mean_shift = np.mean(self.m_shift, axis=0)
                history = f'shift_xy = {mean_shift[0]:.2f}, {mean_shift[1]:.2f}'

        # apply a constant shift
        if constant:

            self.apply_function_to_images(apply_shift,
                                          self.m_image_in_port,
                                          self.m_image_out_port,
                                          'Shifting the images',
                                          func_args=(self.m_shift,
                                                     self.m_interpolation))

            # if self.m_fit_in_port is None or constant:
            history = f'shift_xy = {self.m_shift[0]:.2f}, {self.m_shift[1]:.2f}'

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('ShiftImagesModule', history)
        self.m_image_out_port.close_port()


class WaffleCenteringModule(ProcessingModule):
    """
    Pipeline module for centering of coronagraphic data for which dedicated center frames with
    satellite spots are available.
    """

    __author__ = 'Alexander Bohn'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 center_in_tag: str,
                 image_out_tag: str,
                 size: Optional[float] = None,
                 center: Optional[Tuple[float, float]] = None,
                 radius: float = 45.,
                 pattern: str = None,
                 angle: float = 45.,
                 sigma: float = 0.06,
                 dither: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with science images that are read as input.
        center_in_tag : str
            Tag of the database entry with the center frame that is read as input.
        image_out_tag : str
            Tag of the database entry with the centered images that are written as output. Should
            be different from *image_in_tag*.
        size : float, None
            Image size (arcsec) for both dimensions. Original image size is used if set to None.
        center : tuple(float, float), None
            Approximate position (x0, y0) of the coronagraph. The center of the image is used if
            set to None.
        radius : float
            Approximate separation (pix) of the satellite spots from the star. For IFS data, the
            separation of the spots in the image with the shortest wavelength is required.
        pattern : str, None
            Waffle pattern that is used ('x' or '+'). This parameter will be deprecated in a future
            release. Please use the ``angle`` parameter instead. The parameter will be ignored if
            set to None.
        angle : float
            Angle offset (deg) in clockwise direction of the satellite spots with respect to the
            '+' orientation (i.e. when the spots are located along the horizontal and vertical
            axis). The previous use of the '+' pattern corresponds to 0 degrees and 'x' pattern
            corresponds to 45 degrees. SPHERE/IFS data requires an angle of 55.48 degrees.
        sigma : float
            Standard deviation (arcsec) of the Gaussian kernel that is used for the unsharp
            masking.
        dither : bool
            Apply dithering correction based on the ``DITHER_X`` and ``DITHER_Y`` attributes.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_center_in_port = self.add_input_port(center_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_size = size
        self.m_center = center
        self.m_radius = radius
        self.m_pattern = pattern
        self.m_angle = angle
        self.m_sigma = sigma
        self.m_dither = dither

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Locates the position of the calibration spots in the center
        frame. From the four spots, the position of the star behind the coronagraph is fitted,
        and the images are shifted and cropped.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _get_center(image_number: int,
                        center: Optional[Tuple[int, int]]) -> Tuple[np.ndarray, Tuple[int, int]]:

            if center_shape[-3] > 1:
                warnings.warn('Multiple center images found. Using the first image of the stack.')

            if ndim == 3:
                center_frame = self.m_center_in_port[0, ]
            elif ndim == 4:
                center_frame = self.m_center_in_port[image_number, 0, ]

            if center is None:
                center = center_pixel(center_frame)
            else:
                center = (int(np.floor(center[0])), int(np.floor(center[1])))

            return center_frame, center

        center_shape = self.m_center_in_port.get_shape()
        im_shape = self.m_image_in_port.get_shape()
        ndim = self.m_image_in_port.get_ndim()

        center_frame, self.m_center = _get_center(0, self.m_center)

        # Read in wavelength information or set it to default values
        if ndim == 4:
            wavelength = self.m_image_in_port.get_attribute('WAVELENGTH')

            if wavelength is None:
                raise ValueError('The wavelength information is required to centre IFS data. '
                                 'Please add it via the WavelengthReadingModule before using '
                                 'the WaffleCenteringModule.')

            if im_shape[0] != center_shape[0]:
                raise ValueError(f'Number of science wavelength channels: {im_shape[0]}. '
                                 f'Number of center wavelength channels: {center_shape[0]}. '
                                 'Exactly one center image per wavelength is required.')

            wavelength_min = np.min(wavelength)

        elif ndim == 3:
            # for none ifs data, use default value
            wavelength = [1.]
            wavelength_min = 1.

        # check if science and center images have the same shape
        if im_shape[-2:] != center_shape[-2:]:
            raise ValueError('Science and center images should have the same shape.')

        # Setting angle via pattern (used for backwards compability)
        if self.m_pattern is not None:

            if self.m_pattern == 'x':
                self.m_angle = 45.

            elif self.m_pattern == '+':
                self.m_angle = 0.

            else:
                raise ValueError(f'The pattern {self.m_pattern} is not valid. Please select '
                                 f'either \'x\' or \'+\'.')

            warnings.warn(f'The \'pattern\' parameter will be deprecated in a future release. '
                          f'Please Use the \'angle\' parameter instead and set it to '
                          f'{self.m_angle} degrees.',
                          DeprecationWarning)

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        self.m_sigma /= pixscale

        if self.m_size is not None:
            self.m_size = int(math.ceil(self.m_size/pixscale))

        if self.m_dither:
            dither_x = self.m_image_in_port.get_attribute('DITHER_X')
            dither_y = self.m_image_in_port.get_attribute('DITHER_Y')

            nframes = self.m_image_in_port.get_attribute('NFRAMES')
            nframes = np.cumsum(nframes)
            nframes = np.insert(nframes, 0, 0)

        # size of center image, only works with odd value
        ref_image_size = 21

        # Arrays for the positions
        x_pos = np.zeros(4)
        y_pos = np.zeros(4)

        # Arrays for the center position for each wavelength
        x_center = np.zeros((len(wavelength)))
        y_center = np.zeros((len(wavelength)))

        # Loop for 4 waffle spots
        for w, wave_nr in enumerate(wavelength):

            # Prapre centering frame
            center_frame, _ = _get_center(w, self.m_center)

            center_frame_unsharp = center_frame - gaussian_filter(input=center_frame,
                                                                  sigma=self.m_sigma)

            for i in range(4):
                # Approximate positions of waffle spots
                radius = self.m_radius * wave_nr / wavelength_min

                x_0 = np.floor(self.m_center[0] + radius *
                               np.cos(self.m_angle*np.pi/180 + np.pi / 4. * (2 * i)))

                y_0 = np.floor(self.m_center[1] + radius *
                               np.sin(self.m_angle*np.pi/180 + np.pi / 4. * (2 * i)))

                tmp_center_frame = crop_image(image=center_frame_unsharp,
                                              center=(int(y_0), int(x_0)),
                                              size=ref_image_size)

                # find maximum in tmp image
                coords = np.unravel_index(indices=np.argmax(tmp_center_frame),
                                          shape=tmp_center_frame.shape)

                y_max, x_max = coords[0], coords[1]

                pixmax = tmp_center_frame[y_max, x_max]
                max_pos = np.array([x_max, y_max]).reshape(1, 2)

                # Check whether it is the correct maximum: second brightest pixel should be nearby
                tmp_center_frame[y_max, x_max] = 0.

                # introduce distance parameter
                dist = np.inf

                while dist > 2:
                    coords = np.unravel_index(indices=np.argmax(tmp_center_frame),
                                              shape=tmp_center_frame.shape)

                    y_max_new, x_max_new = coords[0], coords[1]

                    pixmax_new = tmp_center_frame[y_max_new, x_max_new]

                    # Caculate minimal distance to previous points
                    tmp_center_frame[y_max_new, x_max_new] = 0.

                    dist = np.amin(np.linalg.norm(np.vstack((max_pos[:, 0]-x_max_new,
                                                             max_pos[:, 1]-y_max_new)),
                                                  axis=0))

                    if dist <= 2 and pixmax_new < pixmax:
                        break

                    max_pos = np.vstack((max_pos, [x_max_new, y_max_new]))

                    x_max = x_max_new
                    y_max = y_max_new
                    pixmax = pixmax_new

                x_0 = x_0 - (ref_image_size-1)/2 + x_max
                y_0 = y_0 - (ref_image_size-1)/2 + y_max

                # create reference image around determined maximum
                ref_center_frame = crop_image(image=center_frame_unsharp,
                                              center=(int(y_0), int(x_0)),
                                              size=ref_image_size)

                # Fit the data using astropy.modeling
                gauss_init = models.Gaussian2D(amplitude=np.amax(ref_center_frame),
                                               x_mean=x_0,
                                               y_mean=y_0,
                                               x_stddev=1.,
                                               y_stddev=1.,
                                               theta=0.)

                fit_gauss = fitting.LevMarLSQFitter()

                y_grid, x_grid = np.mgrid[y_0-(ref_image_size-1)/2:y_0+(ref_image_size-1)/2+1,
                                          x_0-(ref_image_size-1)/2:x_0+(ref_image_size-1)/2+1]

                gauss = fit_gauss(gauss_init,
                                  x_grid,
                                  y_grid,
                                  ref_center_frame)

                x_pos[i] = gauss.x_mean.value
                y_pos[i] = gauss.y_mean.value

            # Find star position as intersection of two lines

            x_center[w] = ((y_pos[0]-x_pos[0]*(y_pos[2]-y_pos[0])/(x_pos[2]-float(x_pos[0]))) -
                           (y_pos[1]-x_pos[1]*(y_pos[1]-y_pos[3])/(x_pos[1]-float(x_pos[3])))) / \
                          ((y_pos[1]-y_pos[3])/(x_pos[1]-float(x_pos[3])) -
                           (y_pos[2]-y_pos[0])/(x_pos[2]-float(x_pos[0])))

            y_center[w] = x_center[w]*(y_pos[1]-y_pos[3])/(x_pos[1]-float(x_pos[3])) + \
                (y_pos[1]-x_pos[1]*(y_pos[1]-y_pos[3])/(x_pos[1]-float(x_pos[3])))

        # Adjust science images
        nimages = self.m_image_in_port.get_shape()[-3]
        npix = self.m_image_in_port.get_shape()[-2]
        nwavelengths = len(wavelength)

        start_time = time.time()

        for i in range(nimages):
            im_storage = []
            for j in range(nwavelengths):
                im_index = i*nwavelengths + j

                progress(im_index, nimages*nwavelengths, 'Centering the images...', start_time)

                if ndim == 3:
                    image = self.m_image_in_port[i, ]
                elif ndim == 4:
                    image = self.m_image_in_port[j, i, ]

                shift_yx = np.array([(float(im_shape[-2])-1.)/2. - y_center[j],
                                     (float(im_shape[-1])-1.)/2. - x_center[j]])

                if self.m_dither:
                    index = np.digitize(i, nframes, right=False) - 1

                    shift_yx[0] -= dither_y[index]
                    shift_yx[1] -= dither_x[index]

                if npix % 2 == 0 and self.m_size is not None:
                    im_tmp = np.zeros((image.shape[0]+1, image.shape[1]+1))
                    im_tmp[:-1, :-1] = image
                    image = im_tmp

                    shift_yx[0] += 0.5
                    shift_yx[1] += 0.5

                im_shift = shift_image(image, shift_yx, 'spline')

                if self.m_size is not None:
                    im_crop = crop_image(im_shift, None, self.m_size)
                    im_storage.append(im_crop)
                else:
                    im_storage.append(im_shift)

            if ndim == 3:
                self.m_image_out_port.append(im_storage[0], data_dim=3)
            elif ndim == 4:
                self.m_image_out_port.append(np.asarray(im_storage), data_dim=4)

        print(f'Center [x, y] = [{x_center}, {y_center}]')

        history = f'[x, y] = [{round(x_center[j], 2)}, {round(y_center[j], 2)}]'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('WaffleCenteringModule', history)
        self.m_image_out_port.close_port()



"""
Functions for spectral differential imaging.
"""

import numpy as np

from typeguard import typechecked

from pynpoint.util.image import scale_image, shift_image


@typechecked
def sdi_scaling(image_in: np.ndarray,
                scaling: np.ndarray) -> np.ndarray:

    """
    Function to rescale the images by their wavelength ratios.

    Parameters
    ----------
    image_in : np.ndarray
        Data to rescale
    scaling : np.ndarray
        Scaling factors.

    Returns
    -------
    np.ndarray
        Rescaled images with the same shape as ``image_in``.
    """

    if image_in.shape[0] != scaling.shape[0]:
        raise ValueError('The number of wavelengths is not equal to the number of available '
                         'scaling factors.')

    image_out = np.zeros(image_in.shape)

    for i in range(image_in.shape[0]):
        swaps = scale_image(image_in[i, ], scaling[i], scaling[i])

        npix_del = swaps.shape[-1] - image_out.shape[-1]

        if npix_del == 0:
            image_out[i, ] = swaps

        else:
            if npix_del % 2 == 0:
                npix_del_a = int(npix_del/2)
                npix_del_b = int(npix_del/2)

            else:
                npix_del_a = int((npix_del-1)/2)
                npix_del_b = int((npix_del+1)/2)

            image_out[i, ] = swaps[npix_del_a:-npix_del_b, npix_del_a:-npix_del_b]

        if npix_del % 2 == 1:
            image_out[i, ] = shift_image(image_out[i, ], (-0.5, -0.5), interpolation='spline')

    return image_out


@typechecked
def scaling_factors(wavelengths: np.ndarray) -> np.ndarray:
    """
    Function to calculate the scaling factors for SDI.

    Parameters
    ----------
    wavelengths : np.ndarray
        Array with the wavelength of each frame.

    Returns
    -------
    np.ndarray
        Scaling factors.
    """

    return max(wavelengths) / wavelengths



import os
import math

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.processing.badpixel import BadPixelSigmaFilterModule, BadPixelMapModule, \
                                         BadPixelInterpolationModule, BadPixelTimeFilterModule, \
                                         ReplaceBadPixelsModule
from pynpoint.util.tests import create_config, remove_test_data


class TestBadPixel:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        np.random.seed(1)

        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        dark = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        flat = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))

        images[0, 5, 5] = 1.
        dark[:, 5, 5] = 1.
        flat[:, 8, 8] = -1.
        flat[:, 9, 9] = -1.
        flat[:, 10, 10] = -1.

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'w') as hdf_file:
            hdf_file.create_dataset('images', data=images)
            hdf_file.create_dataset('dark', data=dark)
            hdf_file.create_dataset('flat', data=flat)

        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_bad_pixel_sigma_filter(self) -> None:

        module = BadPixelSigmaFilterModule(name_in='sigma1',
                                           image_in_tag='images',
                                           image_out_tag='sigma1',
                                           map_out_tag='None',
                                           box=9,
                                           sigma=3.,
                                           iterate=5)

        self.pipeline.add_module(module)
        self.pipeline.run_module('sigma1')

        data = self.pipeline.get_data('sigma1')
        assert np.sum(data) == pytest.approx(0.006513475520308432, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_map_out(self) -> None:

        module = BadPixelSigmaFilterModule(name_in='sigma2',
                                           image_in_tag='images',
                                           image_out_tag='sigma2',
                                           map_out_tag='bpmap',
                                           box=9,
                                           sigma=2.,
                                           iterate=3)

        self.pipeline.add_module(module)
        self.pipeline.run_module('sigma2')

        data = self.pipeline.get_data('sigma2')
        assert data[0, 0, 0] == pytest.approx(-2.4570591355257687e-05, rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(9.903775276151606e-06, rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(0.011777887008566097, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data('bpmap')
        assert data[0, 1, 1] == pytest.approx(1., rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(0., rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(519.0, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_map(self) -> None:

        module = BadPixelMapModule(name_in='bp_map1',
                                   dark_in_tag='dark',
                                   flat_in_tag='flat',
                                   bp_map_out_tag='bp_map1',
                                   dark_threshold=0.99,
                                   flat_threshold=-0.99)

        self.pipeline.add_module(module)
        self.pipeline.run_module('bp_map1')

        data = self.pipeline.get_data('bp_map1')
        assert data[0, 0, 0] == pytest.approx(1., rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(0., rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(117., rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

    def test_map_no_dark(self) -> None:

        module = BadPixelMapModule(name_in='bp_map2',
                                   dark_in_tag=None,
                                   flat_in_tag='flat',
                                   bp_map_out_tag='bp_map2',
                                   dark_threshold=0.99,
                                   flat_threshold=-0.99)

        self.pipeline.add_module(module)
        self.pipeline.run_module('bp_map2')

        data = self.pipeline.get_data('bp_map2')
        assert np.sum(data) == pytest.approx(118., rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

    def test_map_no_flat(self) -> None:

        module = BadPixelMapModule(name_in='bp_map3',
                                   dark_in_tag='dark',
                                   flat_in_tag=None,
                                   bp_map_out_tag='bp_map3',
                                   dark_threshold=0.99,
                                   flat_threshold=-0.99)

        self.pipeline.add_module(module)
        self.pipeline.run_module('bp_map3')

        data = self.pipeline.get_data('bp_map3')
        assert np.sum(data) == pytest.approx(120., rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

    def test_bad_pixel_interpolation(self) -> None:

        module = BadPixelInterpolationModule(name_in='interpolation',
                                             image_in_tag='images',
                                             bad_pixel_map_tag='bp_map1',
                                             image_out_tag='interpolation',
                                             iterations=10)

        self.pipeline.add_module(module)
        self.pipeline.run_module('interpolation')

        data = self.pipeline.get_data('interpolation')
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=1e-6, abs=0.)
        assert data[0, 5, 5] == pytest.approx(-1.4292408645473845e-05, rel=1e-6, abs=0.)
        assert np.sum(data) == pytest.approx(0.008683344127872174, rel=1e-6, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_time(self) -> None:

        module = BadPixelTimeFilterModule(name_in='time',
                                          image_in_tag='images',
                                          image_out_tag='time',
                                          sigma=(2., 2.))

        self.pipeline.add_module(module)
        self.pipeline.run_module('time')

        data = self.pipeline.get_data('time')
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(-0.00017468532119886812, rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(0.004175672043832705, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_replace_bad_pixels(self) -> None:

        module = ReplaceBadPixelsModule(name_in='replace1',
                                        image_in_tag='images',
                                        map_in_tag='bp_map1',
                                        image_out_tag='replace',
                                        size=2,
                                        replace='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('replace1')

        data = self.pipeline.get_data('replace')
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(4.260114004413933e-05, rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(0.00883357395370896, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        module = ReplaceBadPixelsModule(name_in='replace2',
                                        image_in_tag='images',
                                        map_in_tag='bp_map1',
                                        image_out_tag='replace',
                                        size=2,
                                        replace='median')

        self.pipeline.add_module(module)
        self.pipeline.run_module('replace2')

        data = self.pipeline.get_data('replace')
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)
        assert data[0, 5, 5] == pytest.approx(4.327154179438619e-05, rel=self.limit, abs=0.)
        assert np.sum(data) == pytest.approx(0.008489525337709688, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        module = ReplaceBadPixelsModule(name_in='replace3',
                                        image_in_tag='images',
                                        map_in_tag='bp_map1',
                                        image_out_tag='replace',
                                        size=2,
                                        replace='nan')

        self.pipeline.add_module(module)
        self.pipeline.run_module('replace3')

        data = self.pipeline.get_data('replace')
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)
        assert math.isnan(data[0, 5, 5])
        assert np.nansum(data) == pytest.approx(0.009049653234723834, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)



"""
Functions for PSF subtraction.
"""

from typing import Optional, Union, Tuple

import numpy as np

from scipy.ndimage import rotate
from sklearn.decomposition import PCA
from typeguard import typechecked

from pynpoint.util.image import scale_image, shift_image


@typechecked
def pca_psf_subtraction(images: np.ndarray,
                        angles: Optional[np.ndarray],
                        pca_number: Union[int, np.int32, np.int64],
                        scales: Optional[np.ndarray] = None,
                        pca_sklearn: Optional[PCA] = None,
                        im_shape: Optional[tuple] = None,
                        indices: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Function for PSF subtraction with PCA.

    Parameters
    ----------
    images : np.ndarray
        Stack of images. Also used as reference images if ```pca_sklearn``` is set to None. The
        data should have the original 3D shape if ``pca_sklearn`` is set to None or it should be
        in a 2D reshaped format if ``pca_sklearn`` is not set to None.
    angles : np.ndarray
        Parallactic angles (deg).
    pca_number : int
        Number of principal components.
    scales : np.ndarray, None
        Scaling factors for SDI. Not used if set to None.
    pca_sklearn : sklearn.decomposition.pca.PCA, None
        PCA object with the principal components.
    im_shape : tuple(int, int, int), None
        The original 3D shape of the stack with images. Only required if ``pca_sklearn`` is not set
        to None.
    indices : np.ndarray, None
        Array with the indices of the pixels that are used for the PSF subtraction. All pixels are
        used if set to None.

    Returns
    -------
    np.ndarray
        Residuals of the PSF subtraction.
    np.ndarray
        Derotated residuals of the PSF subtraction.
    """

    if pca_sklearn is None:
        # Create a PCA object if not provided as argument
        pca_sklearn = PCA(n_components=pca_number, svd_solver='arpack')

        # The 3D shape of the array with images
        im_shape = images.shape

        if indices is None:
            # Select the first image and get the unmasked image indices
            im_star = images[0, ].reshape(-1)
            indices = np.where(im_star != 0.)[0]

        # Reshape the images and select the unmasked pixels
        im_reshape = images.reshape(im_shape[0], im_shape[1]*im_shape[2])
        im_reshape = im_reshape[:, indices]

        # Subtract the mean image
        # This is also done by sklearn.decomposition.PCA.fit()
        im_reshape -= np.mean(im_reshape, axis=0)

        # Fit the principal components
        pca_sklearn.fit(im_reshape)

    else:
        # If the PCA object is already there then so are the reshaped data
        im_reshape = np.copy(images)

    # Project the data on the principal components
    # Note that this is the same as sklearn.decomposition.PCA.transform()
    # It is harcoded because the number of components has been adjusted
    pca_rep = np.matmul(pca_sklearn.components_[:pca_number], im_reshape.T)

    # The zeros are added with vstack to account for the components that have not been used for the
    # transformation to the lower-dimensional space, while they were initiated with the PCA object.
    # Since inverse_transform uses the number of initial components, the zeros are added for
    # components > pca_number. These components do not impact the inverse transformation.
    zeros = np.zeros((pca_sklearn.n_components - pca_number, im_reshape.shape[0]))
    pca_rep = np.vstack((pca_rep, zeros)).T

    # Transform the data back to the original space
    psf_model = pca_sklearn.inverse_transform(pca_rep)

    # Create an array with the original shape
    residuals = np.zeros((im_shape[0], im_shape[1]*im_shape[2]))

    # Select all pixel indices if set to None
    if indices is None:
        indices = np.arange(0, im_reshape.shape[1], 1)

    # Subtract the PSF model
    residuals[:, indices] = im_reshape - psf_model

    # Reshape the residuals to the original shape
    residuals = residuals.reshape(im_shape)

    # ----------- back scale images
    scal_cor = np.zeros(residuals.shape)

    if scales is not None:

        # check if the number of parang is equal to the number of images
        if residuals.shape[0] != scales.shape[0]:
            raise ValueError(f'The number of images ({residuals.shape[0]}) is not equal to the '
                             f'number of wavelengths ({scales.shape[0]}).')

        for i, _ in enumerate(scales):
            # rescaling the images
            swaps = scale_image(residuals[i, ], 1/scales[i], 1/scales[i])

            npix_del = scal_cor.shape[-1] - swaps.shape[-1]

            if npix_del == 0:
                scal_cor[i, ] = swaps

            else:
                if npix_del % 2 == 0:
                    npix_del_a = int(npix_del/2)
                    npix_del_b = int(npix_del/2)

                else:
                    npix_del_a = int((npix_del-1)/2)
                    npix_del_b = int((npix_del+1)/2)

                scal_cor[i, npix_del_a:-npix_del_b, npix_del_a:-npix_del_b] = swaps

                if npix_del % 2 == 1:
                    scal_cor[i, ] = shift_image(scal_cor[i, ], (0.5, 0.5), interpolation='spline')

    else:
        scal_cor = residuals

    res_rot = np.zeros(residuals.shape)

    if angles is not None:

        # Check if the number of parang is equal to the number of images
        if residuals.shape[0] != angles.shape[0]:
            raise ValueError(f'The number of images ({residuals.shape[0]}) is not equal to the '
                             f'number of parallactic angles ({angles.shape[0]}).')

        for j, item in enumerate(angles):
            res_rot[j, ] = rotate(scal_cor[j, ], item, reshape=False)

    else:
        res_rot = scal_cor

    return scal_cor, res_rot



import os

from typing import Tuple

import pytest
import numpy as np

from pynpoint.core.dataio import OutputPort, DataStorage, InputPort
from pynpoint.util.tests import create_random


class TestOutputPort:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.storage = DataStorage(os.path.dirname(__file__) + '/PynPoint_database.hdf5')

        create_random(os.path.dirname(__file__))

    def teardown_class(self) -> None:

        os.remove(os.path.dirname(__file__) + '/PynPoint_database.hdf5')

    def create_input_port(self, tag_name: str) -> InputPort:

        inport = InputPort(tag_name, self.storage)
        inport.open_port()

        return inport

    def create_output_port(self, tag_name: str) -> OutputPort:

        outport = OutputPort(tag_name, self.storage)

        return outport

    def test_create_instance(self) -> None:

        with pytest.raises(ValueError) as error:
            OutputPort('config', self.storage)

        assert str(error.value) == 'The tag name \'config\' is reserved for the central ' \
                                   'configuration of PynPoint.'

        with pytest.raises(ValueError) as error:
            OutputPort('fits_header', self.storage)

        assert str(error.value) == 'The tag name \'fits_header\' is reserved for storage of the ' \
                                   'FITS headers.'

        active_port = OutputPort('test', self.storage, activate_init=True)
        deactive_port = OutputPort('test', self.storage, activate_init=False)
        control_port = InputPort('test', self.storage)

        deactive_port.open_port()
        deactive_port.set_all(np.asarray([0, 1, 2, 3]))
        deactive_port.flush()

        with pytest.warns(UserWarning) as warning:
            control_port.get_all()

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'No data under the tag which is linked by the ' \
                                             'InputPort.'

        active_port.set_all(np.asarray([0, 1, 2, 3]))
        active_port.flush()

        assert np.array_equal(np.asarray([0, 1, 2, 3]), control_port.get_all())

        active_port.del_all_data()

    # def test_set_all_new_data(self) -> None:
    #
    #     outport = self.create_output_port('new_data')
    #
    #     # ----- 1D input -----
    #
    #     data = [1, 3]
    #     outport.set_all(data, data_dim=1)
    #
    #     inport = self.create_input_port('new_data')
    #
    #     assert np.array_equal(inport.get_all(), [1., 3.])
    #     outport.del_all_data()
    #
    #     data = [1, 3]
    #     outport.set_all(data, data_dim=2)
    #     assert np.array_equal(inport.get_all(), [[1, 3]])
    #     outport.del_all_data()
    #
    #     # ----- 2D input -----
    #
    #     data = [[1, 3], [2, 4]]
    #     outport.set_all(data, data_dim=2)
    #     assert np.array_equal(inport.get_all(), [[1, 3], [2, 4]])
    #     outport.del_all_data()
    #
    #     data = [[1, 3], [2, 4]]
    #     outport.set_all(data, data_dim=3)
    #     assert np.array_equal(inport.get_all(), [[[1, 3], [2, 4]]])
    #     outport.del_all_data()
    #
    #     # ----- 3D input -----
    #
    #     data = [[[1, 3], [2, 4]], [[1, 3], [2, 4]]]
    #     outport.set_all(data, data_dim=3)
    #     assert np.array_equal(inport.get_all(), [[[1, 3], [2, 4]], [[1, 3], [2, 4]]])
    #     outport.del_all_data()

    # def test_set_all_error(self) -> None:
    #
    #     # ---- Test database not set -----
    #
    #     data = [1, 2, 3, 4, 0]
    #
    #     with pytest.warns(UserWarning) as record:
    #         out_port = OutputPort('some_data')
    #         out_port.set_all(data)
    #
    #     assert len(record) == 1
    #
    #     assert record[0].message.args[0] == 'OutputPort can not store data unless a database is ' \
    #                                         'connected.'
    #
    #     # ---- Test data dim of actual data for new data entry is < 1 or > 5
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     data = [[[[[[2, 2], ], ], ], ]]
    #
    #     with pytest.raises(ValueError) as error:
    #         out_port.set_all(data, data_dim=2)
    #
    #     assert str(error.value) == 'Output port can only save numpy arrays from 1D to 5D. Use ' \
    #                                'Port attributes to save as int, float, or string.'
    #
    #     # ---- Test data dim of data_dim for new data entry is < 1 or > 5
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     data = [1, 2, 4]
    #
    #     with pytest.raises(ValueError) as error:
    #         out_port.set_all(data, data_dim=0)
    #
    #     assert str(error.value) == 'The data dimensions should be 1D, 2D, 3D, 4D, or 5D.'
    #
    #     # ---- Test data_dim for new data entry is smaller than actual data
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     data = [[1], [2]]
    #
    #     with pytest.raises(ValueError) as error:
    #         out_port.set_all(data, data_dim=1)
    #
    #     assert str(error.value) == 'The dimensions of the data should be equal to or larger ' \
    #                                'than the dimensions of the input data.'
    #
    #     # ---- Test data_dim == 3 and actual size == 1
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     data = [1, 2]
    #
    #     with pytest.raises(ValueError) as error:
    #         out_port.set_all(data, data_dim=3)
    #
    #     assert str(error.value) == 'Cannot initialize 1D data in 3D data container.'
    #
    # def test_set_all_keep_attributes(self) -> None:
    #
    #     def init_out_port() -> Tuple[OutputPort, InputPort]:
    #         out_port = self.create_output_port('new_data')
    #         control = self.create_input_port('new_data')
    #
    #         data = [2, 3, 4]
    #         out_port.set_all(data)
    #         out_port.add_attribute(name='test1', value=1)
    #         out_port.add_attribute(name='test2', value=12)
    #
    #         return out_port, control
    #
    #     out_port, control = init_out_port()
    #
    #     with pytest.warns(UserWarning) as warning:
    #         out_port.set_all([[]], data_dim=2, keep_attributes=True)
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'The new dataset that is stored under the tag name ' \
    #                                          '\'new_data\' is empty.'
    #
    #     assert control.get_attribute('test1') == 1
    #     assert control.get_attribute('test2') == 12
    #
    #     out_port.del_all_data()
    #
    # def test_append_new_data(self) -> None:
    #
    #     # using append even if no data exists
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     # ----- 1D input -----
    #
    #     data = [3, ]
    #     out_port.append(data)
    #
    #     control = self.create_input_port('new_data')
    #     assert control.get_all() == data
    #     out_port.del_all_data()
    #
    #     # ----- 2D input -----
    #
    #     data = [[3, 3], [3, 2]]
    #     out_port.append(data)
    #
    #     assert np.array_equal(control.get_all(), data)
    #     out_port.del_all_data()
    #
    #     # ----- 3D input -----
    #
    #     data = [[[3, 3], [3, 2]], [[3, 1], [3, 1]]]
    #     out_port.append(data)
    #
    #     assert np.array_equal(control.get_all(), data)
    #     out_port.del_all_data()
    #
    # def test_append_existing_data(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     # ----- 1D -----
    #
    #     out_port.append([2, 3, 5], data_dim=1)
    #     out_port.append([3, 3, 5])
    #
    #     control = self.create_input_port('new_data')
    #
    #     assert np.array_equal(control.get_all(), [2, 3, 5, 3, 3, 5])
    #
    #     out_port.del_all_data()
    #
    #     # ----- 2D -----
    #     # 1D input append to 1D data
    #
    #     out_port.append([1, 1], data_dim=2)
    #     out_port.append([3, 3])
    #
    #     assert np.array_equal(control.get_all(), [[1, 1], [3, 3]])
    #
    #     out_port.del_all_data()
    #
    #     # 1D input append to 2D data
    #
    #     out_port.append([[2, 3], [1, 1]], data_dim=2)
    #     out_port.append([3, 3])
    #
    #     assert np.array_equal(control.get_all(), [[2, 3], [1, 1], [3, 3]])
    #
    #     out_port.del_all_data()
    #
    #     # 2D input append to 2D data
    #
    #     out_port.append([[2, 3], [1, 1]], data_dim=2)
    #     out_port.append([[3, 3], [8, 8]])
    #
    #     assert np.array_equal(control.get_all(), [[2, 3], [1, 1], [3, 3], [8, 8]])
    #
    #     out_port.del_all_data()
    #
    #     # 2D input append to 3D data
    #
    #     out_port.append([[[2, 3], [1, 1]], [[2, 4], [1, 1]]], data_dim=3)
    #
    #     out_port.append([[3, 3], [8, 8]])
    #
    #     assert np.array_equal(control.get_all(),
    #                           [[[2, 3], [1, 1]],
    #                            [[2, 4], [1, 1]],
    #                            [[3, 3], [8, 8]]])
    #
    #     out_port.del_all_data()
    #
    #     # 3D input append to 3D data
    #
    #     out_port.append([[[2, 3], [1, 1]], [[2, 4], [1, 1]]], data_dim=3)
    #     out_port.append([[[22, 7], [10, 221]], [[223, 46], [1, 15]]])
    #
    #     assert np.array_equal(control.get_all(),
    #                           [[[2, 3], [1, 1]],
    #                            [[2, 4], [1, 1]],
    #                            [[22, 7], [10, 221]],
    #                            [[223, 46], [1, 15]]])
    #
    #     out_port.del_all_data()
    #
    # def test_append_existing_data_force_overwriting(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     # Error case (no force)
    #
    #     out_port.append([2, 3, 5], data_dim=1)
    #     out_port.append([[[22, 7], [10, 221]], [[223, 46], [1, 15]]], force=True)
    #
    #     control = self.create_input_port('new_data')
    #
    #     assert np.array_equal(control.get_all(), [[[22, 7], [10, 221]], [[223, 46], [1, 15]]])
    #
    #     out_port.del_all_data()
    #
    # def test_append_existing_data_error(self) -> None:
    #
    #     # ---- port not active ----
    #     out_port = self.create_output_port('new_data')
    #     out_port.deactivate()
    #
    #     data = [1, ]
    #
    #     out_port.append(data)
    #     out_port.del_all_data()
    #     out_port.activate()
    #
    #     # 1 Element input
    #
    #     # 1D input
    #     # 2D input
    #     # 3D input
    #
    #     # Error case (no force)
    #     out_port.set_all([2, 3, 5], data_dim=1)
    #
    #     with pytest.raises(ValueError) as error:
    #         out_port.append([[[22, 7], [10, 221]], [[223, 46], [1, 15]]])
    #
    #     assert str(error.value) == 'The port tag \'new_data\' is already used with a different ' \
    #                                'data type. The \'force\' parameter can be used to replace ' \
    #                                'the tag.'
    #     out_port.del_all_data()
    #
    # def test_set_data_using_slicing(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     out_port.set_all([2, 5, 6, 7, ])
    #     out_port[3] = 44
    #
    #     control = self.create_input_port('new_data')
    #
    #     assert np.array_equal(control.get_all(), [2, 5, 6, 44, ])
    #
    #     out_port.deactivate()
    #     out_port[2] = 0
    #
    #     assert np.array_equal(control.get_all(), [2, 5, 6, 44, ])
    #
    #     out_port.activate()
    #     out_port.del_all_data()
    #
    # def test_del_all_data(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([0, 1])
    #     out_port.del_all_data()
    #
    #     control = self.create_input_port('new_data')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_all()
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'No data under the tag which is linked by the ' \
    #                                          'InputPort.'
    #
    # def test_add_static_attribute(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([1])
    #     out_port.add_attribute('attr1', value=5)
    #     out_port.add_attribute('attr2', value='no')
    #
    #     control = self.create_input_port('new_data')
    #     assert control.get_attribute('attr1') == 5
    #
    #     out_port.add_attribute('attr1', value=6)
    #     assert control.get_attribute('attr1') == 6
    #     assert control.get_attribute('attr2') == 'no'
    #
    #     out_port.deactivate()
    #     out_port.add_attribute('attr3', value=33)
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_attribute('attr3')
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'The attribute \'attr3\' was not found.'
    #
    #     out_port.activate()
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    # def test_add_static_attribute_error(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     # add attribute while no data is set
    #     with pytest.warns(UserWarning) as warning:
    #         out_port.add_attribute('attr1', value=6)
    #
    #     # check that only one warning was raised
    #     assert len(warning) == 1
    #
    #     # check that the message matches
    #     assert warning[0].message.args[0] == 'Can not store the attribute \'attr1\' because ' \
    #                                          'the dataset \'new_data\' does not exist.'
    #
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    # def test_add_non_static_attribute(self) -> None:
    #
    #     # two different data types
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([1])
    #     out_port.add_attribute('attr1', value=[6, 3], static=False)
    #
    #     control = self.create_input_port('new_data')
    #     assert np.array_equal(control.get_attribute('attr1'), [6, 3])
    #
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    # def test_append_attribute_data(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.del_all_data()
    #     out_port.set_all([1])
    #     out_port.add_attribute('attr1', value=[2, 3], static=False)
    #
    #     control = self.create_input_port('new_data')
    #     assert np.array_equal(control.get_attribute('attr1'), [2, 3])
    #
    #     out_port.append_attribute_data('attr1', value=2)
    #     assert np.array_equal(control.get_attribute('attr1'), [2, 3, 2])
    #
    #     out_port.deactivate()
    #     out_port.append_attribute_data('attr1', value=2)
    #     assert np.array_equal(control.get_attribute('attr1'), [2, 3, 2])
    #
    #     out_port.activate()
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    # def test_copy_attributes(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #     out_port.set_all([0, ])
    #
    #     # some static attributes
    #
    #     out_port.add_attribute('attr1', 33)
    #     out_port.add_attribute('attr2', 'string')
    #     out_port.add_attribute('attr3', 3.141)
    #
    #     # non static attributes
    #
    #     out_port.add_attribute('attr_non_static', [3, 4, 5, 6], static=False)
    #
    #     copy_port = self.create_output_port('other_data')
    #     copy_port.del_all_attributes()
    #     copy_port.del_all_data()
    #
    #     copy_port.set_all([1, ])
    #
    #     # for attribute overwriting
    #
    #     copy_port.add_attribute('attr_non_static', [3, 4, 44, 6], static=False)
    #
    #     control = self.create_input_port('new_data')
    #     copy_port.copy_attributes(control)
    #
    #     copy_control = self.create_input_port('other_data')
    #
    #     assert copy_control.get_attribute('attr1') == 33
    #     assert copy_control.get_attribute('attr2') == 'string'
    #
    #     assert np.array_equal(copy_control.get_attribute('attr3'), 3.141)
    #     assert np.array_equal(copy_control.get_attribute('attr_non_static'), [3, 4, 5, 6])
    #
    #     copy_port.del_all_attributes()
    #     copy_port.del_all_data()
    #
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    #     port = self.create_output_port('test')
    #     port.deactivate()
    #
    #     assert port.copy_attributes(control) is None
    #
    #     port = self.create_input_port('test')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         port.get_all_non_static_attributes()
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'No data under the tag which is linked by the ' \
    #                                          'InputPort.'
    #
    # def test_copy_attributes_same_tag(self) -> None:
    #
    #     out_port1 = self.create_output_port('new_data')
    #     out_port1.set_all([0, ])
    #
    #     out_port2 = self.create_output_port('new_data')
    #     out_port2.set_all([2, ])
    #
    #     out_port1.add_attribute('attr1', 2)
    #
    #     control1 = self.create_input_port('new_data')
    #     out_port2.copy_attributes(control1)
    #
    #     control2 = self.create_input_port('new_data')
    #     assert control2.get_attribute('attr1') == 2
    #
    #     out_port1.del_all_data()
    #     out_port1.del_all_attributes()
    #
    # def test_del_attribute(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([0, ])
    #
    #     # static
    #
    #     out_port.add_attribute('attr1', 4)
    #     out_port.add_attribute('attr2', 5)
    #
    #     # non static
    #
    #     out_port.add_attribute('attr_non_static_1', [1, 2, 3], static=False)
    #     out_port.add_attribute('attr_non_static_2', [2, 4, 6, 8], static=False)
    #     out_port.del_attribute('attr1')
    #     out_port.del_attribute('attr_non_static_1')
    #
    #     # check is only the chosen attributes are deleted and the rest is still there
    #
    #     control = self.create_input_port('new_data')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_attribute('attr1')
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'The attribute \'attr1\' was not found.'
    #
    #     assert control.get_attribute('attr2') == 5
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_attribute('attr_non_static_1')
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'The attribute \'attr_non_static_1\' was not found.'
    #
    #     assert np.array_equal(control.get_attribute('attr_non_static_2'), [2, 4, 6, 8])
    #
    #     out_port.del_all_data()
    #     out_port.del_all_attributes()
    #
    # def test_del_attribute_error_case(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([0, ])
    #
    #     # deactivated port
    #
    #     out_port.add_attribute('attr_1', 5.554)
    #     out_port.deactivate()
    #     out_port.del_attribute('attr_1')
    #
    #     control = self.create_input_port('new_data')
    #     assert control.get_attribute('attr_1') == 5.554
    #
    #     out_port.activate()
    #
    #     # not existing
    #     with pytest.warns(UserWarning) as warning:
    #         out_port.del_attribute('not_existing')
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'Attribute \'not_existing\' does not exist and ' \
    #                                          'could not be deleted.'
    #
    #     out_port.del_all_attributes()
    #     out_port.del_all_data()
    #
    # def test_del_all_attributes(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([0, ])
    #     out_port.add_attribute('attr_1', 4)
    #     out_port.add_attribute('attr_2', [1, 3], static=False)
    #     out_port.del_all_attributes()
    #
    #     control = self.create_input_port('new_data')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_attribute('attr_1')
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'The attribute \'attr_1\' was not found.'
    #
    #     with pytest.warns(UserWarning) as warning:
    #         control.get_attribute('attr_2')
    #
    #     assert len(warning) == 1
    #     assert warning[0].message.args[0] == 'The attribute \'attr_2\' was not found.'
    #
    #     out_port.del_all_data()
    #
    # def test_add_history(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #     out_port.set_all([0, ])
    #     out_port.add_history('Test', 'history')
    #
    #     control = self.create_input_port('new_data')
    #     assert control.get_attribute('History: Test') == 'history'
    #
    # def test_check_attribute(self) -> None:
    #
    #     out_port = self.create_output_port('new_data')
    #
    #     out_port.set_all([0, ])
    #     out_port.add_attribute('static', 5, static=True)
    #     out_port.add_attribute('non-static', np.arange(1, 11, 1), static=False)
    #
    #     assert out_port.check_static_attribute('static', 5) == 0
    #     assert out_port.check_static_attribute('test', 3) == 1
    #     assert out_port.check_static_attribute('static', 33) == -1
    #
    #     assert out_port.check_non_static_attribute('non-static', np.arange(1, 11, 1)) == 0
    #     assert out_port.check_non_static_attribute('test', np.arange(1, 11, 1)) == 1
    #     assert out_port.check_non_static_attribute('non-static', np.arange(10, 21, 1)) == -1
    #
    #     out_port.deactivate()
    #
    #     assert out_port.check_static_attribute('static', 5) is None
    #     assert out_port.check_non_static_attribute('non-static', np.arange(1, 11, 1)) is None
    #
    #     out_port.activate()
    #     out_port.del_all_data()
    #     out_port.del_all_attributes()
    #
    # def test_append_empty_data_new(self) -> None:
    #
    #     out_port = self.create_output_port('empty_data')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         out_port.append([])
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'The new dataset that is stored under the tag name ' \
    #                                          '\'empty_data\' is empty.'
    #
    # def test_append_empty_data_add(self) -> None:
    #
    #     out_port = self.create_output_port('empty_data')
    #
    #     with pytest.warns(UserWarning) as warning:
    #         out_port.append([])
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'The dataset that is appended under the tag name ' \
    #                                          '\'empty_data\' is empty.'



"""
Functions for MCMC sampling.
"""

import math

from typing import Optional, Tuple

import numpy as np

from typeguard import typechecked

from pynpoint.util.analysis import fake_planet, merit_function
from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.residuals import combine_residuals


@typechecked
def lnprob(param: np.ndarray,
           bounds: Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]],
           images: np.ndarray,
           psf: np.ndarray,
           mask: np.ndarray,
           parang: np.ndarray,
           psf_scaling: float,
           pixscale: float,
           pca_number: int,
           extra_rot: float,
           aperture: Tuple[int, int, float],
           indices: np.ndarray,
           merit: str,
           residuals: str,
           var_noise: Optional[float]) -> float:
    """
    Function for the log posterior function. Should be placed at the highest level of the
    Python module to be pickable for the multiprocessing.

    Parameters
    ----------
    param : numpy.ndarray
        The separation (arcsec), angle (deg), and contrast (mag). The angle is measured in
        counterclockwise direction with respect to the positive y-axis.
    bounds : tuple(tuple(float, float), tuple(float, float), tuple(float, float))
        The boundaries of the separation (arcsec), angle (deg), and contrast (mag). Each set of
        boundaries is specified as a tuple.
    images : numpy.ndarray
        Stack with images.
    psf : numpy.ndarray
        PSF template, either a single image (2D) or a cube (3D) with the dimensions equal to
        *images*.
    mask : numpy.ndarray
        Array with the circular mask (zeros) of the central and outer regions.
    parang : numpy.ndarray
        Array with the angles for derotation.
    psf_scaling : float
        Additional scaling factor of the planet flux (e.g., to correct for a neutral density
        filter). Should be negative in order to inject negative fake planets.
    pixscale : float
        Additional scaling factor of the planet flux (e.g., to correct for a neutral density
        filter). Should be negative in order to inject negative fake planets.
    pca_number : int
        Number of principal components used for the PSF subtraction.
    extra_rot : float
        Additional rotation angle of the images (deg).
    aperture : tuple(int, int, float)
        Position (y, x) of the aperture center (pix) and aperture radius (pix).
    indices : numpy.ndarray
        Non-masked image indices.
    merit : str
        Figure of merit that is used for the likelihood function ('gaussian' or 'poisson').
        Pixels are assumed to be independent measurements which are expected to be equal to
        zero in case the best-fit negative PSF template is injected. With 'gaussian', the
        variance is estimated from the pixel values within an annulus at the separation of
        the aperture (but excluding the pixels within the aperture). With 'poisson', a
        Poisson distribution is assumed for the variance of each pixel value.
    residuals : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    var_noise : float, None
        Variance of the noise which is required when `merit` is set to 'gaussian' or 'hessian'.

    Returns
    -------
    float
        Log posterior probability.
    """

    @typechecked
    def _lnprior() -> float:
        """
        Internal function for the log prior function.

        Returns
        -------
        float
            Log prior.
        """

        if bounds[0][0] <= param[0] <= bounds[0][1] and \
           bounds[1][0] <= param[1] <= bounds[1][1] and \
           bounds[2][0] <= param[2] <= bounds[2][1]:

            ln_prior = 0.

        else:

            ln_prior = -np.inf

        return ln_prior

    @typechecked
    def _lnlike() -> float:
        """
        Internal function for the log likelihood function.

        Returns
        -------
        float
            Log likelihood.
        """

        sep, ang, mag = param

        fake = fake_planet(images=images,
                           psf=psf,
                           parang=parang-extra_rot,
                           position=(sep/pixscale, ang),
                           magnitude=mag,
                           psf_scaling=psf_scaling)

        im_res_rot, im_res_derot = pca_psf_subtraction(images=fake*mask,
                                                       angles=-1.*parang+extra_rot,
                                                       pca_number=pca_number,
                                                       indices=indices)

        res_stack = combine_residuals(method=residuals,
                                      res_rot=im_res_derot,
                                      residuals=im_res_rot,
                                      angles=parang)

        chi_square = merit_function(residuals=res_stack[0, ],
                                    merit=merit,
                                    aperture=aperture,
                                    sigma=0.,
                                    var_noise=var_noise)

        return -0.5*chi_square

    ln_prior = _lnprior()

    if math.isinf(ln_prior):
        ln_prob = -np.inf
    else:
        ln_prob = ln_prior + _lnlike()

    return ln_prob



"""
Additional custom types to make type hints easier to read.
"""

from typing import Union

import numpy as np


StaticAttribute = Union[str, float, int, np.generic]
NonStaticAttribute = Union[np.ndarray, tuple, list]



"""
Wrapper utils for the wavelet functions for the mlpy cwt implementation (see continous.py)
"""

import numpy as np

from numba import jit
from typeguard import typechecked
from scipy.special import gamma, hermite
from scipy.signal import medfilt
from statsmodels.robust import mad

from pynpoint.util.continuous import autoscales, cwt, icwt
# from pynpoint.util.continuous import fourier_from_scales


# This function cannot by @typechecked because of a compatibility issue with numba
@jit(cache=True, nopython=True)
def _fast_zeros(soft: bool,
                spectrum: np.ndarray,
                uthresh: float) -> np.ndarray:
    """
    Fast numba method to modify values in the wavelet space by using a hard or soft threshold
    function.

    Parameters
    ----------
    soft : bool
        If True soft the threshold function will be used, otherwise a hard threshold is applied.
    spectrum : numpy.ndarray
        The input 2D wavelet space.
    uthresh : float
        Threshold used by the threshold function.

    Returns
    -------
    numpy.ndarray
        Modified spectrum.
    """

    if soft:
        for i in range(0, spectrum.shape[0], 1):
            for j in range(0, spectrum.shape[1], 1):
                tmp_value = spectrum[i, j].real
                if abs(spectrum[i, j]) > uthresh:
                    spectrum[i, j] = np.sign(tmp_value) * (abs(tmp_value) - uthresh)
                else:
                    spectrum[i, j] = 0
    else:
        for i in range(0, spectrum.shape[0], 1):
            for j in range(0, spectrum.shape[1], 1):
                if abs(spectrum[i, j]) < uthresh:
                    spectrum[i, j] = 0

    return spectrum


class WaveletAnalysisCapsule:
    """
    Capsule class to process one 1d time series using the CWT and wavelet de-nosing by wavelet
    shrinkage.
    """

    @typechecked
    def __init__(self,
                 signal_in: np.ndarray,
                 wavelet_in: str = 'dog',
                 order: int = 2,
                 padding: str = 'none',
                 frequency_resolution: float = 0.5) -> None:
        """
        Parameters
        ----------
        signal_in : numpy.ndarray
            1D input signal.
        wavelet_in : str
            Wavelet function ('dog' or 'morlet').
        order : int
            Order of the wavelet function.
        padding : str
            Padding method ('zero', 'mirror', or 'none').
        frequency_resolution : float
            Wavelet space resolution in scale/frequency.

        Returns
        -------
        NoneType
            None
        """

        # save input data
        self.m_supported_wavelets = ['dog', 'morlet']

        # check supported wavelets
        if wavelet_in not in self.m_supported_wavelets:
            raise ValueError(f'Wavelet {wavelet_in} is not supported')

        if wavelet_in == 'dog':
            self._m_c_reconstructions = {2: 3.5987,
                                         4: 2.4014,
                                         6: 1.9212,
                                         8: 1.6467,
                                         12: 1.3307,
                                         16: 1.1464,
                                         20: 1.0222,
                                         30: 0.8312,
                                         40: 0.7183,
                                         60: 0.5853}
        elif wavelet_in == 'morlet':
            self._m_c_reconstructions = {5: 0.9484,
                                         6: 0.7784,
                                         7: 0.6616,
                                         8: 0.5758,
                                         10: 0.4579,
                                         12: 0.3804,
                                         14: 0.3254,
                                         16: 0.2844,
                                         20: 0.2272}
        self.m_wavelet = wavelet_in

        if padding not in ['none', 'zero', 'mirror']:
            raise ValueError('Padding can only be none, zero or mirror')

        self._m_data = signal_in - np.ones(len(signal_in)) * np.mean(signal_in)
        self.m_padding = padding
        self.__pad_signal()
        self._m_data_size = len(self._m_data)
        self._m_data_mean = np.mean(signal_in)

        if order not in self._m_c_reconstructions:
            raise ValueError('Wavelet ' + str(wavelet_in) + ' does not support order '
                             + str(order) + ". \n Only orders: " +
                             str(sorted(self._m_c_reconstructions.keys())).strip('[]') +
                             " are supported")
        self.m_order = order
        self._m_c_final_reconstruction = self._m_c_reconstructions[order]

        # create scales for wavelet transform
        self._m_scales = autoscales(N=self._m_data_size,
                                    dt=1,
                                    dj=frequency_resolution,
                                    wf=wavelet_in,
                                    p=order)

        self._m_number_of_scales = len(self._m_scales)
        self._m_frequency_resolution = frequency_resolution

        self.m_spectrum = None

    # --- functions for reconstruction value
    @staticmethod
    @typechecked
    def _morlet_function(omega0: float,
                         x_in: float) -> np.complex128:
        """
        Returns
        -------
        numpy.complex128
            Morlet function.
        """

        return np.pi**(-0.25) * np.exp(1j * omega0 * x_in) * np.exp(-x_in**2/2.0)

    @staticmethod
    @typechecked
    def _dog_function(order: int,
                      x_in: float) -> float:
        """
        Returns
        -------
        float
            DOG function.
        """

        p_hpoly = hermite(order)[int(x_in / np.power(2, 0.5))]
        herm = p_hpoly / (np.power(2, float(order) / 2))

        return ((-1)**(order+1)) / np.sqrt(gamma(order + 0.5)) * herm

    @typechecked
    def __pad_signal(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        padding_length = int(len(self._m_data) * 0.5)

        if self.m_padding == 'zero':
            new_data = np.append(self._m_data, np.zeros(padding_length, dtype=np.float64))
            self._m_data = np.append(np.zeros(padding_length, dtype=np.float64), new_data)

        elif self.m_padding == 'mirror':
            left_half_signal = self._m_data[:padding_length]
            right_half_signal = self._m_data[padding_length:]
            new_data = np.append(self._m_data, right_half_signal[::-1])
            self._m_data = np.append(left_half_signal[::-1], new_data)

    @typechecked
    def __compute_reconstruction_factor(self) -> float:
        """
        Computes the reconstruction factor.

        Returns
        -------
        float
            Reconstruction factor.
        """

        freq_res = self._m_frequency_resolution
        wavelet = self.m_wavelet
        order = self.m_order

        if wavelet == 'morlet':
            zero_function = self._morlet_function(order, 0)
        else:
            zero_function = self._dog_function(order, 0)

        c_delta = self._m_c_final_reconstruction

        reconstruction_factor = freq_res/(c_delta * zero_function)

        return reconstruction_factor.real

    @typechecked
    def compute_cwt(self) -> None:
        """
        Compute the wavelet space of the given input signal.

        Returns
        -------
        NoneType
            None
        """

        self.m_spectrum = cwt(self._m_data,
                              dt=1,
                              scales=self._m_scales,
                              wf=self.m_wavelet,
                              p=self.m_order)

    @typechecked
    def update_signal(self) -> None:
        """
        Updates the internal signal by the reconstruction of the current wavelet space.

        Returns
        -------
        NoneType
            None
        """

        self._m_data = icwt(self.m_spectrum, scales=self._m_scales)
        reconstruction_factor = self.__compute_reconstruction_factor()
        self._m_data *= reconstruction_factor

    @typechecked
    def denoise_spectrum(self,
                         soft: bool = False) -> None:
        """
        Applies wavelet shrinkage on the current wavelet space (m_spectrum) by either a hard of
        soft threshold function.

        Parameters
        ----------
        soft : bool
            If True a soft threshold is used, hard otherwise.

        Returns
        -------
        NoneType
            None
        """

        if self.m_padding != 'none':
            noise_length_4 = len(self._m_data) // 4
            noise_spectrum = self.m_spectrum[0, noise_length_4: (noise_length_4 * 3)].real

        else:
            noise_spectrum = self.m_spectrum[0, :].real

        sigma = mad(noise_spectrum)
        uthresh = sigma*np.sqrt(2.0*np.log(len(noise_spectrum)))

        self.m_spectrum = _fast_zeros(soft, self.m_spectrum, uthresh)

    @typechecked
    def median_filter(self) -> None:
        """
        Applies a median filter on the internal 1d signal. Can be useful for cosmic ray correction
        after temporal de-noising

        Returns
        -------
        NoneType
            None
        """

        self._m_data = medfilt(self._m_data, 19)

    @typechecked
    def get_signal(self) -> np.ndarray:
        """
        Returns the current version of the 1d signal. Use update_signal() in advance in order to get
        the current reconstruction of the wavelet space. Removes padded values as well.

        Returns
        -------
        numpy.ndarray
            Current version of the 1D signal.
        """

        tmp_data = self._m_data + np.ones(len(self._m_data)) * self._m_data_mean

        if self.m_padding == 'none':
            return tmp_data

        return tmp_data[len(self._m_data) // 4: 3 * (len(self._m_data) // 4)]

    # def __transform_period(self,
    #                        period):
    #
    #     tmp_y = fourier_from_scales(self._m_scales,
    #                                 self.m_wavelet,
    #                                 self.m_order)
    #
    #     def __transformation(x):
    #         return np.log2(x + 1) * tmp_y[-1] / np.log2(tmp_y[-1] + 1)
    #
    #     cutoff_scaled = __transformation(period)
    #
    #     scale_new = tmp_y[-1] - tmp_y[0]
    #     scale_old = self.m_spectrum.shape[0]
    #
    #     factor = scale_old / scale_new
    #     cutoff_scaled *= factor
    #
    #     return cutoff_scaled

    # ----- plotting functions --------

    # def __plot_or_save_spectrum(self):
    #     plt.close()
    #
    #     plt.figure(figsize=(8, 6))
    #     plt.subplot(1, 1, 1)
    #
    #     tmp_y = fourier_from_scales(self._m_scales,
    #                                 self.m_wavelet,
    #                                 self.m_order)
    #
    #     tmp_x = np.arange(0, self._m_data_size + 1, 1)
    #
    #     scaled_spec = copy.deepcopy(self.m_spectrum.real)
    #     for i, _ in enumerate(scaled_spec):
    #         scaled_spec[i] /= np.sqrt(self._m_scales[i])
    #
    #     plt.imshow(abs(scaled_spec),
    #                aspect='auto',
    #                extent=[tmp_x[0],
    #                        tmp_x[-1],
    #                        tmp_y[0],
    #                        tmp_y[-1]],
    #                cmap=plt.get_cmap("gist_ncar"),
    #                origin='lower')
    #
    #     # COI first part (only for DOG) with padding
    #
    #     inner_frequency = 2.*np.pi/np.sqrt(self.m_order + 0.5)
    #     coi = np.append(np.zeros(len(tmp_x)/4),
    #                     tmp_x[0:len(tmp_x) / 4])
    #     coi = np.append(coi,
    #                     tmp_x[0:len(tmp_x) / 4][::-1])
    #     coi = np.append(coi,
    #                     np.zeros(len(tmp_x) / 4))
    #
    #     plt.plot(np.arange(0, len(coi), 1.0),
    #              inner_frequency * coi / np.sqrt(2),
    #              color="white")
    #
    #     plt.ylim([tmp_y[0],
    #               tmp_y[-1]])
    #
    #     plt.fill_between(np.arange(0, len(coi), 1.0),
    #                      inner_frequency * coi / np.sqrt(2),
    #                      np.ones(len(coi)) * tmp_y[-1],
    #                      facecolor="none",
    #                      edgecolor='white',
    #                      alpha=0.4,
    #                      hatch="x")
    #
    #     plt.yscale('log', basey=2)
    #     plt.ylabel("Period in [s]")
    #     plt.xlabel("Time in [s]")
    #     plt.title("Spectrum computed with CWT using '" + str(self.m_wavelet) +
    #               "' wavelet of order " + str(self.m_order))
    #
    # def plot_spectrum(self):
    #     """
    #     Shows a plot of the current wavelet space.
    #     :return: None
    #     """
    #
    #     self.__plot_or_save_spectrum()
    #     plt.show()
    #
    # def save_spectrum(self,
    #                   location):
    #     """
    #     Saves a plot of the current wavelet space to a given location.
    #     :param location: Save location
    #     :type location: str
    #     :return: None
    #     """
    #     self.__plot_or_save_spectrum()
    #     plt.savefig(location)
    #     plt.close()
    #
    # def __plot_or_save_signal(self):
    #     plt.close()
    #     plt.plot(self._m_data)
    #     plt.title("Signal")
    #     plt.ylabel("Value of the function")
    #     plt.xlim([0, self._m_data_size])
    #     plt.xlabel("Time in [s]")
    #
    # def plot_signal(self):
    #     """
    #     Plot the current signal.
    #     :return: None
    #     """
    #     self.__plot_or_save_signal()
    #     plt.show()
    #
    # def save_signal(self,
    #                 location):
    #     """
    #     Saves a plot of the current signal to a given location.
    #     :param location: Save location
    #     :type location: str
    #     :return: None
    #     """
    #     self.__plot_or_save_signal()
    #     plt.savefig(location)



"""
Functions for image processing.
"""

import math

from typing import Optional, Tuple, Union

import numpy as np

from scipy.ndimage import fourier_shift, shift, rotate
from skimage.transform import rescale
from typeguard import typechecked


@typechecked
def center_pixel(image: np.ndarray) -> Tuple[int, int]:
    """
    Function to get the pixel position of the image center. Note that this position
    can not be unambiguously defined for an even-sized image. Python indexing starts
    at 0 so the coordinates of the pixel in the bottom-left corner are (0, 0).

    Parameters
    ----------
    image : np.ndarray
        Input image (2D or 3D).

    Returns
    -------
    tuple(int, int)
        Pixel position (y, x) of the image center.
    """

    if image.shape[-2] % 2 == 0 and image.shape[-1] % 2 == 0:
        center = (image.shape[-2] // 2 - 1, image.shape[-1] // 2 - 1)

    elif image.shape[-2] % 2 == 0 and image.shape[-1] % 2 == 1:
        center = (image.shape[-2] // 2 - 1, (image.shape[-1]-1) // 2)

    elif image.shape[-2] % 2 == 1 and image.shape[-1] % 2 == 0:
        center = ((image.shape[-2] - 1) // 2, image.shape[-1] // 2 - 1)

    elif image.shape[-2] % 2 == 1 and image.shape[-1] % 2 == 1:
        center = ((image.shape[-2] - 1) // 2, (image.shape[-1] - 1) // 2)

    else:
        raise RuntimeError('Unexpected image shape. This error should not occur.')

    return center


@typechecked
def center_subpixel(image: np.ndarray) -> Tuple[float, float]:
    """
    Function to get the precise position of the image center. The center of the pixel in the
    bottom left corner of the image is defined as (0, 0), so the bottom left corner of the
    image is located at (-0.5, -0.5).

    Parameters
    ----------
    image : np.ndarray
        Input image (2D or 3D).

    Returns
    -------
    tuple(float, float)
        Subpixel position (y, x) of the image center.
    """

    center_x = float(image.shape[-1]) / 2 - 0.5
    center_y = float(image.shape[-2]) / 2 - 0.5

    return center_y, center_x


@typechecked
def crop_image(image: np.ndarray,
               center: Optional[tuple],
               size: int,
               copy: bool = True) -> np.ndarray:
    """
    Function to crop square images around a specified position.

    Parameters
    ----------
    image : np.ndarray
        Input image (2D or 3D).
    center : tuple(int, int), None
        The new image center (y, x). The center of the image is used if set to None.
    size : int
        Image size (pix) for both dimensions. Increased by 1 pixel if size is an even number.
    copy : bool
        Whether or not to return a copy (instead of a view) of the cropped image (default: True).

    Returns
    -------
    np.ndarray
        Cropped odd-sized image (2D or 3D).
    """

    if center is None or (center[0] is None and center[1] is None):
        center = center_pixel(image)

        # if image.shape[-1] % 2 == 0:
        #     warnings.warn('The image is even-size so there is not a uniquely defined pixel in '
        #                   'the center of the image. The image center is determined (with pixel '
        #                   'precision) with the pynpoint.util.image.center_pixel function.')

    if size % 2 == 0:
        size += 1

    x_start = center[1] - (size - 1) // 2
    x_end = center[1] + (size - 1) // 2 + 1

    y_start = center[0] - (size - 1) // 2
    y_end = center[0] + (size - 1) // 2 + 1

    if x_start < 0 or y_start < 0 or x_end > image.shape[-1] or y_end > image.shape[-2]:
        raise ValueError('Target image resolution does not fit inside the input image resolution.')

    return np.array(image[..., y_start:y_end, x_start:x_end], copy=copy)


@typechecked
def rotate_images(images: np.ndarray,
                  angles: np.ndarray) -> np.ndarray:
    """
    Function to rotate all images in clockwise direction.

    Parameters
    ----------
    images : np.ndarray
        Stack of images (3D).
    angles : np.ndarray
        Rotation angles (deg).

    Returns
    -------
    np.ndarray
        Rotated images.
    """

    im_rot = np.zeros(images.shape)

    for i, item in enumerate(angles):
        im_rot[i, ] = rotate(input=images[i, ], angle=item, reshape=False)

    return im_rot


@typechecked
def create_mask(im_shape: Tuple[int, int],
                size: Union[Tuple[float, float],
                            Tuple[float, None],
                            Tuple[None, float],
                            Tuple[None, None]]) -> np.ndarray:
    """
    Function to create a mask for the central and outer image regions.

    Parameters
    ----------
    im_shape : tuple(int, int)
        Image size in both dimensions.
    size : tuple(float, float)
        Size (pix) of the inner and outer mask.

    Returns
    -------
    np.ndarray
        Image mask.
    """

    mask = np.ones(im_shape)
    npix = im_shape[0]

    if size[0] is not None or size[1] is not None:

        if npix % 2 == 0:
            x_grid = y_grid = np.linspace(-npix / 2 + 0.5, npix / 2 - 0.5, npix)
        else:
            x_grid = y_grid = np.linspace(-(npix - 1) / 2, (npix - 1) / 2, npix)

        xx_grid, yy_grid = np.meshgrid(x_grid, y_grid)
        rr_grid = np.sqrt(xx_grid**2 + yy_grid**2)

        if size[0] is not None:
            mask[rr_grid < size[0]] = 0.

        if size[1] is not None:
            if size[1] > npix / 2:
                size = (size[0], npix / 2)
            mask[rr_grid > size[1]] = 0.

    return mask


@typechecked
def shift_image(image: np.ndarray,
                shift_yx: Union[Tuple[float, float], np.ndarray],
                interpolation: str,
                mode: str = 'constant') -> np.ndarray:
    """
    Function to shift an image.

    Parameters
    ----------
    image : np.ndarray
        Input image (2D or 3D). If 3D the image is not shifted along the 0th axis.
    shift_yx : tuple(float, float), np.ndarray
        Shift (y, x) to be applied (pix). An additional shift of zero pixels will be added
        for the first dimension in case the input image is 3D.
    interpolation : str
        Interpolation type ('spline', 'bilinear', or 'fft').
    mode : str
        Interpolation mode.

    Returns
    -------
    np.ndarray
        Shifted image.
    """

    if image.ndim == 2:
        shift_val = (shift_yx[0], shift_yx[1])
    elif image.ndim == 3:
        shift_val = (0, shift_yx[0], shift_yx[1])
    else:
        raise ValueError('Invalid number of dimensions for image: must be 2 or 3')

    if interpolation == 'spline':
        im_center = shift(image, shift_val, order=5, mode=mode)

    elif interpolation == 'bilinear':
        im_center = shift(image, shift_val, order=1, mode=mode)

    elif interpolation == 'fft':
        fft_shift = fourier_shift(np.fft.fftn(image), shift_val)
        im_center = np.fft.ifftn(fft_shift).real

    else:
        raise ValueError('interpolation must be one of the following: spline, bilinear, fft')

    return im_center


@typechecked
def scale_image(image: np.ndarray,
                scaling_y: Union[float, np.float32],
                scaling_x: Union[float, np.float32]) -> np.ndarray:
    """
    Function to spatially scale an image.

    Parameters
    ----------
    image : np.ndarray
        Input image (2D).
    scaling_y : float
        Scaling factor y.
    scaling_x : float
        Scaling factor x.

    Returns
    -------
    np.ndarray
        Shifted image (2D).
    """

    sum_before = np.sum(image)

    im_scale = rescale(image,
                       (scaling_y, scaling_x),
                       order=5,
                       mode='reflect',
                       channel_axis=None,
                       anti_aliasing=True)

    sum_after = np.sum(im_scale)

    return im_scale * (sum_before / sum_after)


@typechecked
def cartesian_to_polar(center: Tuple[float, float],
                       y_pos: float,
                       x_pos: float) -> Tuple[float, float]:
    """
    Function to convert pixel coordinates to polar coordinates.

    Parameters
    ----------
    center : tuple(float, float)
        Image center (y, x) from :func:`~pynpoint.util.image.center_subpixel`.
    y_pos : float
        Pixel coordinate along the vertical axis. The bottom left corner of the image is
        (-0.5, -0.5).
    x_pos : float
        Pixel coordinate along the horizontal axis. The bottom left corner of the image is
        (-0.5, -0.5).

    Returns
    -------
    tuple(float, float)
        Separation (pix) and position angle (deg). The angle is measured counterclockwise with
        respect to the positive y-axis.
    """

    sep = math.sqrt((center[1] - x_pos)**2 + (center[0] - y_pos)**2)
    ang = math.atan2(y_pos-center[1], x_pos-center[0])
    ang = (math.degrees(ang) - 90) % 360

    return sep, ang


@typechecked
def polar_to_cartesian(image: np.ndarray,
                       sep: float,
                       ang: float) -> Tuple[float, float]:
    """
    Function to convert polar coordinates to pixel coordinates.

    Parameters
    ----------
    image : np.ndarray
        Input image (2D or 3D).
    sep : float
        Separation (pixels).
    ang : float
        Position angle (deg), measured counterclockwise with respect to the positive y-axis.

    Returns
    -------
    tuple(float, float)
        Cartesian coordinates (y, x). The bottom left corner of the image is (-0.5, -0.5).
    """

    center = center_subpixel(image)  # (y, x)

    x_pos = center[1] + sep * math.cos(math.radians(ang + 90))
    y_pos = center[0] + sep * math.sin(math.radians(ang + 90))

    return y_pos, x_pos


@typechecked
def pixel_distance(im_shape: Tuple[int, int],
                   position: Optional[Tuple[int, int]] = None) -> Tuple[
                       np.ndarray, np.ndarray, np.ndarray]:
    """
    Function to calculate the distance of each pixel with respect to a given pixel position.
    Supports both odd and even sized images.

    Parameters
    ----------
    im_shape : tuple(int, int)
        Image shape (y, x).
    position : tuple(int, int)
        Pixel center (y, x) from which the distance is calculated. The image center is used if set
        to None. Python indexing starts at zero so the center of the bottom left pixel is (0, 0).

    Returns
    -------
    np.ndarray
        2D array with the distances of each pixel from the provided pixel position.
    np.ndarray
        2D array with the x coordinates.
    np.ndarray
        2D array with the y coordinates.
    """

    if im_shape[0] % 2 == 0:
        y_grid = np.linspace(-im_shape[0] / 2 + 0.5, im_shape[0] / 2 - 0.5, im_shape[0])

    else:
        y_grid = np.linspace(-(im_shape[0] - 1) / 2, (im_shape[0] - 1) / 2, im_shape[0])

    if im_shape[1] % 2 == 0:
        x_grid = np.linspace(-im_shape[1] / 2 + 0.5, im_shape[1] / 2 - 0.5, im_shape[1])

    else:
        x_grid = np.linspace(-(im_shape[1] - 1) / 2, (im_shape[1] - 1) / 2, im_shape[1])

    if position is not None:
        y_shift = y_grid[position[0]]
        x_shift = x_grid[position[1]]

        y_grid -= y_shift
        x_grid -= x_shift

    xx_grid, yy_grid = np.meshgrid(x_grid, y_grid)

    return np.sqrt(xx_grid**2 + yy_grid**2), xx_grid, yy_grid


@typechecked
def subpixel_distance(im_shape: Tuple[int, int],
                      position: Tuple[float, float],
                      shift_center: bool = True) -> np.ndarray:
    """
    Function to calculate the distance of each pixel with respect to a given subpixel position.
    Supports both odd and even sized images.

    Parameters
    ----------
    im_shape : tuple(int, int)
        Image shape (y, x).
    position : tuple(float, float)
        Pixel center (y, x) from which the distance is calculated. Python indexing starts at zero
        so the bottom left image corner is (-0.5, -0.5).
    shift_center : bool
        Apply the coordinate correction for the image center.

    Returns
    -------
    np.ndarray
        2D array with the distances of each pixel from the provided pixel position.
    """

    # Get 2D x and y coordinates with respect to the image center
    _, xx_grid, yy_grid = pixel_distance(im_shape, position=None)

    if im_shape[0] % 2 == 0:
        # Distance from the image center to the center of the outermost pixel
        # Even sized images
        y_size = im_shape[0] / 2 + 0.5
        x_size = im_shape[1] / 2 + 0.5

    else:
        # Distance from the image center to the center of the outermost pixel
        # Odd sized images
        y_size = (im_shape[0] - 1) / 2
        x_size = (im_shape[1] - 1) / 2

    if shift_center:
        # Shift the image center to the center of the bottom left pixel
        yy_grid += y_size
        xx_grid += x_size

    # Apply a subpixel shift of the coordinate system to the requested position
    yy_grid -= position[0]
    xx_grid -= position[1]

    return np.sqrt(xx_grid**2 + yy_grid**2)


@typechecked
def select_annulus(image_in: np.ndarray,
                   radius_in: float,
                   radius_out: float,
                   mask_position: Optional[Tuple[float, float]] = None,
                   mask_radius: Optional[float] = None) -> np.ndarray:
    """
    image_in : np.ndarray
        Input image.
    radius_in : float
        Inner radius of the annulus (pix).
    radius_out : float
        Outer radius of the annulus (pix).
    mask_position : tuple(float, float), None
        Center (pix) position (y, x) in of the circular region that is excluded. Not used
        if set to None.
    mask_radius : float, None
        Radius (pix) of the circular region that is excluded. Not used if set to None.
    """

    im_shape = image_in.shape

    if im_shape[0] % 2 == 0:
        y_grid = np.linspace(-im_shape[0] / 2 + 0.5, im_shape[0] / 2 - 0.5, im_shape[0])
    else:
        y_grid = np.linspace(-(im_shape[0] - 1) / 2, (im_shape[0] - 1) / 2, im_shape[0])

    if im_shape[1] % 2 == 0:
        x_grid = np.linspace(-im_shape[1] / 2 + 0.5, im_shape[1] / 2 - 0.5, im_shape[1])
    else:
        x_grid = np.linspace(-(im_shape[1] - 1) / 2, (im_shape[1] - 1) / 2, im_shape[1])

    xx_grid, yy_grid = np.meshgrid(x_grid, y_grid)
    rr_grid = np.sqrt(xx_grid**2 + yy_grid**2)

    mask = np.ones(im_shape)

    indices = np.where((rr_grid < radius_in) | (rr_grid > radius_out))
    mask[indices[0], indices[1]] = 0.

    if mask_position is not None and mask_radius is not None:
        distance = subpixel_distance(im_shape=im_shape, position=mask_position)
        indices = np.where(distance < mask_radius)
        mask[indices[0], indices[1]] = 0.

    indices = np.where(mask == 1.)

    return image_in[indices[0], indices[1]]


@typechecked
def rotate_coordinates(center: Tuple[float, float],
                       position: Union[Tuple[float, float], np.ndarray],
                       angle: float) -> Tuple[float, float]:
    """
    Function to rotate coordinates around the image center.

    Parameters
    ----------
    center : tuple(float, float)
        Image center (y, x) with subpixel accuracy.
    position : tuple(float, float)
        Position (y, x) in the image, or a 2D numpy array of positions.
    angle : float
        Angle (deg) to rotate in counterclockwise direction.

    Returns
    -------
    tuple(float, float)
        New position (y, x).
    """

    pos_y = (position[1] - center[1]) * math.sin(np.radians(angle)) + \
            (position[0] - center[0]) * math.cos(np.radians(angle))

    pos_x = (position[1] - center[1]) * math.cos(np.radians(angle)) - \
            (position[0] - center[0]) * math.sin(np.radians(angle))

    return center[0]+pos_y, center[1]+pos_x



"""
Modules for writing data as text file.
"""

import os

from typing import Optional

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import WritingModule


class AttributeWritingModule(WritingModule):
    """
    Module for writing a 1D or 2D array of non-static attributes to a text file.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 attribute: str,
                 file_name: str = 'attributes.dat',
                 output_dir: Optional[str] = None,
                 header: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry from which the ``PARANG`` attribute is read.
        attribute : str
            Name of the non-static attribute as given in the central database (e.g., 'INDEX' or
            'STAR_POSITION').
        file_name : str
            Name of the output file.
        output_dir : str, None
            Output directory where the text file will be stored. If no path is specified then the
            Pypeline default output location is used.
        header : str, None
            Header that is written at the top of the text file.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, output_dir=output_dir)

        self.m_data_port = self.add_input_port(data_tag)

        self.m_file_name = file_name
        self.m_attribute = attribute
        self.m_header = header

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Writes the non-static attributes (1D or 2D) to a a text file.

        Returns
        -------
        NoneType
            None
        """

        if self.m_header is None:
            self.m_header = ''

        print('Writing attribute data...', end='')

        out_name = os.path.join(self.m_output_location, self.m_file_name)

        if self.m_attribute not in self.m_data_port.get_all_non_static_attributes():
            raise ValueError(f'The \'{self.m_attribute}\' attribute is not present in '
                             f'\'{self.m_data_port.tag}\'.')

        values = self.m_data_port.get_attribute(self.m_attribute)

        np.savetxt(out_name, values, header=self.m_header, comments='# ')

        print(' [DONE]')

        self.m_data_port.close_port()


class ParangWritingModule(WritingModule):
    """
    Module for writing a list of parallactic angles to a text file.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str = 'parang.dat',
                 output_dir: Optional[str] = None,
                 header: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry from which the ``PARANG`` attribute is read.
        file_name : str
            Name of the output file.
        output_dir : str, None
            Output directory where the text file will be stored. If no path is specified then the
            Pypeline default output location is used.
        header : str, None
            Header that is written at the top of the text file.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, output_dir=output_dir)

        self.m_data_port = self.add_input_port(data_tag)

        self.m_file_name = file_name
        self.m_header = header

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Writes the parallactic angles from the ``PARANG`` attribute of
        the specified database tag to a a text file.

        Returns
        -------
        NoneType
            None
        """

        print('Writing parallactic angles...', end='')

        if self.m_header is None:
            self.m_header = ''

        out_name = os.path.join(self.m_output_location, self.m_file_name)

        if 'PARANG' not in self.m_data_port.get_all_non_static_attributes():
            raise ValueError(f'The PARANG attribute is not present in \'{self.m_data_port.tag}\'.')

        parang = self.m_data_port.get_attribute('PARANG')

        np.savetxt(out_name, parang, header=self.m_header, comments='# ')

        print(' [DONE]')

        self.m_data_port.close_port()



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfpreparation import PSFpreparationModule
from pynpoint.processing.psfsubtraction import PcaPsfSubtractionModule, ClassicalADIModule
from pynpoint.util.tests import create_config, create_fake_data, remove_test_data


class TestPsfSubtractionAdi:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_fake_data(self.test_dir+'science')
        create_fake_data(self.test_dir+'reference')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['science', 'reference'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='science',
                                   input_dir=self.test_dir+'science')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('science')
        assert np.sum(data) == pytest.approx(11.012854046962481, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        self.pipeline.set_attribute('science', 'PARANG', np.linspace(0., 180., 10), static=False)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='reference',
                                   input_dir=self.test_dir+'reference')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('reference')
        assert np.sum(data) == pytest.approx(11.012854046962481, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

    def test_psf_preparation(self) -> None:

        module = PSFpreparationModule(name_in='prep1',
                                      image_in_tag='science',
                                      image_out_tag='science_prep',
                                      mask_out_tag=None,
                                      norm=False,
                                      resize=None,
                                      cent_size=0.05,
                                      edge_size=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep1')

        data = self.pipeline.get_data('science_prep')
        assert np.sum(data) == pytest.approx(5.029285028467547, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        module = PSFpreparationModule(name_in='prep2',
                                      image_in_tag='reference',
                                      image_out_tag='reference_prep',
                                      mask_out_tag=None,
                                      norm=False,
                                      resize=None,
                                      cent_size=0.05,
                                      edge_size=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep2')

        data = self.pipeline.get_data('reference_prep')
        assert np.sum(data) == pytest.approx(5.029285028467547, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

    def test_classical_adi(self) -> None:

        module = ClassicalADIModule(threshold=None,
                                    nreference=None,
                                    residuals='mean',
                                    extra_rot=0.,
                                    name_in='cadi1',
                                    image_in_tag='science_prep',
                                    res_out_tag='cadi_res',
                                    stack_out_tag='cadi_stack')

        self.pipeline.add_module(module)
        self.pipeline.run_module('cadi1')

        data = self.pipeline.get_data('cadi_res')
        assert np.sum(data) == pytest.approx(0.8381625719865213, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data('cadi_stack')
        assert np.sum(data) == pytest.approx(0.08395606034388256, rel=self.limit, abs=0.)
        assert data.shape == (1, 21, 21)

    def test_classical_adi_threshold(self) -> None:

        module = ClassicalADIModule(threshold=(0.1, 0.03, 1.),
                                    nreference=5,
                                    residuals='median',
                                    extra_rot=0.,
                                    name_in='cadi2',
                                    image_in_tag='science_prep',
                                    res_out_tag='cadi_res',
                                    stack_out_tag='cadi_stack')

        self.pipeline.add_module(module)
        self.pipeline.run_module('cadi2')

        data = self.pipeline.get_data('cadi_res')
        assert np.sum(data) == pytest.approx(0.7158207863548083, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data('cadi_stack')
        assert np.sum(data) == pytest.approx(0.07448334552227256, rel=self.limit, abs=0.)
        assert data.shape == (1, 21, 21)

    def test_psf_subtraction_pca_single(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_single',
                                         images_in_tag='science',
                                         reference_in_tag='science',
                                         res_mean_tag='res_mean_single',
                                         res_median_tag='res_median_single',
                                         res_weighted_tag='res_weighted_single',
                                         res_rot_mean_clip_tag='res_clip_single',
                                         res_arr_out_tag='res_arr_single',
                                         basis_out_tag='basis_single',
                                         extra_rot=45.,
                                         subtract_mean=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_single')

        data = self.pipeline.get_data('res_mean_single')
        assert np.sum(data) == pytest.approx(-0.00011857022709778602, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_median_single')
        assert np.sum(data) == pytest.approx(-0.002184868916566093, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_weighted_single')
        assert np.sum(data) == pytest.approx(0.08102176735226937, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        # data = self.pipeline.get_data('res_clip_single')
        # assert np.sum(data) == pytest.approx(7.09495495339349e-05, rel=self.limit, abs=0.)
        # assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_arr_single1')
        assert np.sum(data) == pytest.approx(-0.0002751385418691618, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data('basis_single')
        assert np.sum(data) == pytest.approx(0.09438697731322143, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_no_mean(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_no_mean',
                                         images_in_tag='science',
                                         reference_in_tag='science',
                                         res_mean_tag='res_mean_no_mean',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_no_mean',
                                         extra_rot=0.,
                                         subtract_mean=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_no_mean')

        data = self.pipeline.get_data('res_mean_no_mean')
        assert np.sum(data) == pytest.approx(0.0006081272007585688, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_no_mean')
        assert np.sum(data) == pytest.approx(5.118005177367776, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_ref',
                                         images_in_tag='science',
                                         reference_in_tag='reference',
                                         res_mean_tag='res_mean_ref',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_ref',
                                         extra_rot=0.,
                                         subtract_mean=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_ref')

        data = self.pipeline.get_data('res_mean_ref')
        assert np.sum(data) == pytest.approx(0.0006330226118859073, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_ref')
        assert np.sum(data) == pytest.approx(0.0943869773132221, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_no_mean(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_ref_no_mean',
                                         images_in_tag='science',
                                         reference_in_tag='reference',
                                         res_mean_tag='res_mean_ref_no_mean',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_ref_no_mean',
                                         extra_rot=0.,
                                         subtract_mean=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_ref_no_mean')

        data = self.pipeline.get_data('res_mean_ref_no_mean')
        assert np.sum(data) == pytest.approx(0.0006081272007585764, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_ref_no_mean')
        assert np.sum(data) == pytest.approx(5.118005177367774, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_pca_single_mask(self) -> None:

        pca = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                      name_in='pca_single_mask',
                                      images_in_tag='science_prep',
                                      reference_in_tag='science_prep',
                                      res_mean_tag='res_mean_single_mask',
                                      res_median_tag='res_median_single_mask',
                                      res_weighted_tag='res_weighted_single_mask',
                                      res_rot_mean_clip_tag='res_clip_single_mask',
                                      res_arr_out_tag='res_arr_single_mask',
                                      basis_out_tag='basis_single_mask',
                                      extra_rot=45.,
                                      subtract_mean=True)

        self.pipeline.add_module(pca)
        self.pipeline.run_module('pca_single_mask')

        data = self.pipeline.get_data('res_mean_single_mask')
        assert np.sum(data) == pytest.approx(0.00010696166038626307, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_median_single_mask')
        assert np.sum(data) == pytest.approx(-0.0021005307611346156, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_weighted_single_mask')
        assert np.sum(data) == pytest.approx(0.06014309988789256, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_clip_single_mask')
        # assert np.sum(data) == pytest.approx(9.35120662148806e-05, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('res_arr_single_mask1')
        assert np.sum(data) == pytest.approx(0.0006170872862547557, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data('basis_single_mask')
        assert np.sum(data) == pytest.approx(0.08411251293842359, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_no_mean_mask(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_no_mean_mask',
                                         images_in_tag='science_prep',
                                         reference_in_tag='science_prep',
                                         res_mean_tag='res_mean_no_mean_mask',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_no_mean_mask',
                                         extra_rot=0.,
                                         subtract_mean=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_no_mean_mask')

        data = self.pipeline.get_data('res_mean_no_mean_mask')
        assert np.sum(data) == pytest.approx(2.3542359949502915e-05, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_no_mean_mask')
        assert np.sum(data) == pytest.approx(5.655460951633232, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_mask(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_ref_mask',
                                         images_in_tag='science_prep',
                                         reference_in_tag='reference_prep',
                                         res_mean_tag='res_mean_ref_mask',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_ref_mask',
                                         extra_rot=0.,
                                         subtract_mean=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_ref_mask')

        data = self.pipeline.get_data('res_mean_ref_mask')
        assert np.sum(data) == pytest.approx(9.400558926815758e-06, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_ref_mask')
        assert np.sum(data) == pytest.approx(0.08411251293842326, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_no_mean_mask(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_ref_no_mean_mask',
                                         images_in_tag='science_prep',
                                         reference_in_tag='reference_prep',
                                         res_mean_tag='res_mean_ref_no_mean_mask',
                                         res_median_tag=None,
                                         res_weighted_tag=None,
                                         res_rot_mean_clip_tag=None,
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_ref_no_mean_mask',
                                         extra_rot=0.,
                                         subtract_mean=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_ref_no_mean_mask')

        data = self.pipeline.get_data('res_mean_ref_no_mean_mask')
        assert np.sum(data) == pytest.approx(2.354235994950671e-05, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data('basis_ref_no_mean_mask')
        assert np.sum(data) == pytest.approx(5.655460951633233, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_pca_multi(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_multi',
                                         images_in_tag='science',
                                         reference_in_tag='science',
                                         res_mean_tag='res_mean_multi',
                                         res_median_tag='res_median_multi',
                                         res_weighted_tag='res_weighted_multi',
                                         res_rot_mean_clip_tag='res_clip_multi',
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_multi',
                                         extra_rot=45.,
                                         subtract_mean=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_multi')

        data_single = self.pipeline.get_data('res_mean_single')
        data_multi = self.pipeline.get_data('res_mean_multi')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('res_median_single')
        data_multi = self.pipeline.get_data('res_median_multi')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('res_weighted_single')
        data_multi = self.pipeline.get_data('res_weighted_multi')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('basis_single')
        data_multi = self.pipeline.get_data('basis_multi')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

    def test_psf_subtraction_pca_multi_mask(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        module = PcaPsfSubtractionModule(pca_numbers=range(1, 3),
                                         name_in='pca_multi_mask',
                                         images_in_tag='science_prep',
                                         reference_in_tag='science_prep',
                                         res_mean_tag='res_mean_multi_mask',
                                         res_median_tag='res_median_multi_mask',
                                         res_weighted_tag='res_weighted_multi_mask',
                                         res_rot_mean_clip_tag='res_clip_multi_mask',
                                         res_arr_out_tag=None,
                                         basis_out_tag='basis_multi_mask',
                                         extra_rot=45.,
                                         subtract_mean=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_multi_mask')

        data_single = self.pipeline.get_data('res_mean_single_mask')
        data_multi = self.pipeline.get_data('res_mean_multi_mask')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('res_median_single_mask')
        data_multi = self.pipeline.get_data('res_median_multi_mask')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('res_weighted_single_mask')
        data_multi = self.pipeline.get_data('res_weighted_multi_mask')
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == \
            pytest.approx(data_multi[data_multi > 1e-12], rel=self.limit, abs=0.)

        data_single = self.pipeline.get_data('basis_single_mask')
        data_multi = self.pipeline.get_data('basis_multi_mask')
        assert data_single.shape == data_multi.shape
        assert data_single == pytest.approx(data_multi, rel=self.limit, abs=0.)

    def test_psf_subtraction_len_parang(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 1

        parang = self.pipeline.get_data('header_science/PARANG')
        self.pipeline.set_attribute('science_prep', 'PARANG', np.append(parang, 0.), static=False)

        module = PcaPsfSubtractionModule(pca_numbers=[1, ],
                                         name_in='pca_len_parang',
                                         images_in_tag='science_prep',
                                         reference_in_tag='science_prep',
                                         res_mean_tag='res_mean_len_parang',
                                         extra_rot=0.)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('pca_len_parang')

        assert str(error.value) == 'The number of images (10) is not equal to the number of ' \
                                   'parallactic angles (11).'



import os

import pytest

from pynpoint.core.pypeline import Pypeline
from pynpoint.core.dataio import ConfigPort, DataStorage
from pynpoint.util.tests import create_config, remove_test_data


class TestConfigPort:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'
        create_config(self.test_dir+'PynPoint_config.ini')

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_create_config_port(self) -> None:

        storage = DataStorage(self.test_dir + 'PynPoint_database.hdf5')

        with pytest.raises(ValueError) as error:
            ConfigPort('images', storage)

        assert str(error.value) == 'The tag name of the central configuration should be ' \
                                   '\'config\'.'

        port = ConfigPort('config', None)

        with pytest.warns(UserWarning) as warning:
            check_error = port._check_error_cases()

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'ConfigPort can not load data unless a database is ' \
                                             'connected.'

        assert not check_error

        port = ConfigPort('config', storage)
        assert isinstance(port, ConfigPort)

        with pytest.warns(UserWarning) as warning:
            port._check_error_cases()

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'No data under the tag which is linked by the ' \
                                             'ConfigPort.'

    def test_get_config_attribute(self) -> None:

        create_config(self.test_dir+'PynPoint_config.ini')
        Pypeline(self.test_dir, self.test_dir, self.test_dir)

        storage = DataStorage(self.test_dir + 'PynPoint_database.hdf5')
        port = ConfigPort('config', None)

        with pytest.warns(UserWarning) as warning:
            attribute = port.get_attribute('CPU')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'ConfigPort can not load data unless a database is ' \
                                             'connected.'

        assert attribute is None

        port = ConfigPort('config', storage)

        attribute = port.get_attribute('CPU')
        assert attribute == 1

        attribute = port.get_attribute('NFRAMES')
        assert attribute == 'NAXIS3'

        attribute = port.get_attribute('PIXSCALE')
        assert attribute == pytest.approx(0.027, rel=self.limit, abs=0.)

        with pytest.warns(UserWarning) as warning:
            attribute = port.get_attribute('test')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'The attribute \'test\' was not found.'

        assert attribute is None



"""
Pipeline modules for spatial filtering of images.
"""

import math
import time

from typeguard import typechecked

from scipy.ndimage import gaussian_filter

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import memory_frames, progress


class GaussianFilterModule(ProcessingModule):
    """
    Pipeline module for applying a Gaussian filter.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 fwhm: float = 1.) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : tuple(str, str)
            Tuple with two tags of the database entry that are read as input.
        image_out_tag : str
            Tag of the database entry with the subtracted images that are written as output.
        fwhm : float
            Full width at half maximum (arcsec) of the Gaussian kernel.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_fwhm = fwhm

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Applies a Gaussian filter to the spatial dimensions of the
        images.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        pixscale = self._m_config_port.get_attribute('PIXSCALE')

        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        sigma = (self.m_fwhm/pixscale) / (2.*math.sqrt(2.*math.log(2.)))  # [pix]

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Applying Gaussian filter...', start_time)

            images = self.m_image_in_port[frames[i]:frames[i+1], ]
            im_filter = gaussian_filter(images, (0, sigma, sigma))

            self.m_image_out_port.append(im_filter, data_dim=3)

        history = f'fwhm [arcsec] = {self.m_fwhm}'
        self.m_image_out_port.add_history('GaussianFilterModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()
